# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, division

import errno
import re

import msgfy
import six
import subprocrunner as spr
import typepy
from humanreadable import ParameterError

from ._common import find_bin_path, logging_context, run_command_helper, validate_within_min_max
from ._const import (
    LIST_MANGLE_TABLE_OPTION,
    ShapingAlgorithm,
    Tc,
    TcCommandOutput,
    TcSubCommand,
    TrafficDirection,
)
from ._error import NetworkInterfaceNotFoundError
from ._iptables import IptablesMangleController, get_iptables_base_command
from ._logger import logger
from ._network import sanitize_network, verify_network_interface
from ._shaping_rule_finder import TcShapingRuleFinder
from ._tc_command_helper import get_tc_base_command
from .shaper.htb import HtbShaper
from .shaper.tbf import TbfShaper


class TrafficControl(object):
    __MIN_PORT = 0
    __MAX_PORT = 65535

    REGEXP_FILE_EXISTS = re.compile("RTNETLINK answers: File exists")

    EXISTS_MSG_TEMPLATE = "\n".join(
        [
            "{:s} try to execute with: ",
            "  (a) --overwrite option if you want to overwrite the existing rule.",
            "  (b) --add option if you want to add a new rule in addition to the existing rules.",
            "  (c) --change option if you want to change the existing rule parameter.",
        ]
    )

    @property
    def device(self):
        return self.__device

    @property
    def tc_target(self):
        return self.__device

    @property
    def ifb_device(self):
        return "ifb{:d}".format(self.__qdisc_major_id)

    @property
    def direction(self):
        return self.__direction

    @property
    def netem_param(self):
        return self.__netem_param

    @property
    def dst_network(self):
        return self.__dst_network

    @property
    def exclude_dst_network(self):
        return self.__exclude_dst_network

    @property
    def src_network(self):
        return self.__src_network

    @property
    def exclude_src_network(self):
        return self.__exclude_src_network

    @property
    def src_port(self):
        return self.__src_port

    @property
    def exclude_src_port(self):
        return self.__exclude_src_port

    @property
    def dst_port(self):
        return self.__dst_port

    @property
    def exclude_dst_port(self):
        return self.__exclude_dst_port

    @property
    def is_change_shaping_rule(self):
        return self.__is_change_shaping_rule

    @property
    def is_add_shaping_rule(self):
        return self.__is_add_shaping_rule

    @property
    def is_enable_iptables(self):
        return self.__is_enable_iptables

    @property
    def qdisc_major_id(self):
        return self.__qdisc_major_id

    @property
    def qdisc_major_id_str(self):
        return "{:x}".format(self.__qdisc_major_id)

    @property
    def ip_version(self):
        return 6 if self.__is_ipv6 else 4

    @property
    def protocol(self):
        return "ipv6" if self.__is_ipv6 else "ip"

    @property
    def protocol_match(self):
        return "ip6" if self.__is_ipv6 else "ip"

    @property
    def tc_command_output(self):
        return self.__tc_command_output

    @property
    def iptables_ctrl(self):
        return self.__iptables_ctrl

    def __init__(
        self,
        device,
        direction=None,
        netem_param=None,
        dst_network=None,
        exclude_dst_network=None,
        src_network=None,
        exclude_src_network=None,
        dst_port=None,
        exclude_dst_port=None,
        src_port=None,
        exclude_src_port=None,
        is_ipv6=False,
        is_change_shaping_rule=False,
        is_add_shaping_rule=False,
        is_enable_iptables=False,
        shaping_algorithm=None,
        tc_command_output=TcCommandOutput.NOT_SET,
    ):
        self.__device = device

        self.__direction = direction
        self.__netem_param = netem_param
        self.__dst_network = dst_network
        self.__exclude_dst_network = exclude_dst_network
        self.__src_network = src_network
        self.__exclude_src_network = exclude_src_network
        self.__src_port = src_port
        self.__exclude_src_port = exclude_src_port
        self.__dst_port = dst_port
        self.__exclude_dst_port = exclude_dst_port
        self.__is_ipv6 = is_ipv6
        self.__is_change_shaping_rule = is_change_shaping_rule
        self.__is_add_shaping_rule = is_add_shaping_rule
        self.__is_enable_iptables = is_enable_iptables
        self.__tc_command_output = tc_command_output

        self.__qdisc_major_id = self.__get_device_qdisc_major_id()

        self.__iptables_ctrl = IptablesMangleController(is_enable_iptables, self.ip_version)

        self.__init_shaper(shaping_algorithm)

    def validate(self):
        verify_network_interface(self.device, self.__tc_command_output)
        self.__netem_param.validate_netem_parameter()
        self.__validate_src_network()
        self.__validate_port()

    def __validate_src_network(self):
        if any(
            [
                typepy.is_null_string(self.src_network),
                self.__shaper.algorithm_name == ShapingAlgorithm.HTB,
            ]
        ):
            return

        if not self.is_enable_iptables:
            raise ParameterError(
                "--iptables option required to use --src-network option",
                value=self.is_enable_iptables,
            )

    def sanitize(self):
        self.__dst_network = sanitize_network(self.dst_network, self.ip_version)
        self.__src_network = sanitize_network(self.src_network, self.ip_version)

    def get_tc_device(self):
        """
        Return a device name that associated network communication direction.
        """

        if self.direction == TrafficDirection.OUTGOING:
            return self.device

        if self.direction == TrafficDirection.INCOMING:
            return self.ifb_device

        raise ParameterError(
            "unknown direction", expected=TrafficDirection.LIST, value=self.direction
        )

    def get_tc_command(self, subcommand):
        return "{:s} {:s}".format(
            get_tc_base_command(subcommand), "change" if self.is_change_shaping_rule else "add"
        )

    def get_command_history(self):
        def tc_command_filter(command):
            if get_iptables_base_command():
                if re.search(
                    "^{:s} {:s}".format(
                        get_iptables_base_command(), re.escape(LIST_MANGLE_TABLE_OPTION)
                    ),
                    command,
                ):
                    return False

            if re.search("^{:s} .* show dev".format(find_bin_path("tc")), command):
                return False

            if re.search("^ip (netns exec |link show )", command):
                return False

            if find_bin_path("getcap"):
                if re.search("^{:s}".format(find_bin_path("getcap")), command):
                    return False

            return True

        return filter(tc_command_filter, spr.SubprocessRunner.get_history())

    def make_srcdst_text(self):
        return "".join(
            [
                six.text_type(item)
                for item in [
                    self.dst_network,
                    self.exclude_dst_network,
                    self.src_network,
                    self.exclude_src_network,
                    self.src_port,
                    self.exclude_src_port,
                    self.dst_port,
                    self.exclude_dst_port,
                ]
            ]
        )

    def set_shaping_rule(self):
        rule_finder = TcShapingRuleFinder(logger=logger, tc=self)
        if self.__is_change_shaping_rule and self.tc_command_output == TcCommandOutput.NOT_SET:
            if rule_finder.find_filter_param() is not None:
                self.__is_change_shaping_rule = True
            else:
                self.__is_change_shaping_rule = False
                self.__is_add_shaping_rule = True

        self.__setup_ifb()

        return self.__shaper.set_shaping()

    def delete_all_rules(self):
        result_list = []

        result_list.append(self.__delete_qdisc())
        result_list.append(self.__delete_ingress_qdisc())

        try:
            result_list.append(self.__delete_ifb_device() == 0)
        except NetworkInterfaceNotFoundError as e:
            logger.debug(msgfy.to_debug_message(e))
            result_list.append(False)

        with logging_context("delete iptables mangle table entries"):
            try:
                self.iptables_ctrl.clear()
            except OSError as e:
                logger.warn("{} (can not delete iptables entries)".format(e))

        return any(result_list)

    def delete_tc(self):
        """
        Delete a specific shaping rule.
        """

        rule_finder = TcShapingRuleFinder(logger=logger, tc=self)
        filter_param = rule_finder.find_filter_param()

        if not filter_param:
            message = "shaping rule not found ({}).".format(rule_finder.get_filter_string())
            if rule_finder.is_empty_filter_condition():
                message += " you can delete all of the shaping rules with --all option."
            logger.error(message)

            return 1

        logger.info("delete a shaping rule: {}".format(dict(filter_param)))

        filter_del_command = (
            "{command:s} del dev {dev:s} protocol {protocol:s} "
            "parent {parent:} handle {handle:s} prio {prio:} u32".format(
                command=get_tc_base_command(TcSubCommand.FILTER),
                dev=rule_finder.get_parsed_device(),
                protocol=filter_param.get(Tc.Param.PROTOCOL),
                parent="{:s}:".format(rule_finder.find_parent().split(":")[0]),
                handle=filter_param.get(Tc.Param.FILTER_ID),
                prio=filter_param.get(Tc.Param.PRIORITY),
            )
        )

        result = run_command_helper(
            command=filter_del_command, ignore_error_msg_regexp=None, notice_msg=None
        )

        rule_finder.clear()
        if not rule_finder.is_any_filter():
            logger.debug("there are no filters remain. delete qdiscs.")
            self.delete_all_rules()

        return result

    def __init_shaper(self, shaping_algorithm):
        if shaping_algorithm is None:
            self.__shaper = None
            return

        if shaping_algorithm == ShapingAlgorithm.HTB:
            self.__shaper = HtbShaper(self)
            return

        if shaping_algorithm == ShapingAlgorithm.TBF:
            self.__shaper = TbfShaper(self)
            return

        raise ParameterError(
            "unknown shaping algorithm", expected=ShapingAlgorithm.LIST, value=shaping_algorithm
        )

    def __validate_port(self):
        validate_within_min_max(
            "src_port", self.src_port, self.__MIN_PORT, self.__MAX_PORT, unit=None
        )

        validate_within_min_max(
            "dst_port", self.dst_port, self.__MIN_PORT, self.__MAX_PORT, unit=None
        )

    def __get_device_qdisc_major_id(self):
        import hashlib

        base_device_hash = hashlib.md5(six.b(self.device)).hexdigest()[:3]
        device_hash_prefix = "1"

        return int(device_hash_prefix + base_device_hash, 16)

    def __setup_ifb(self):
        if self.direction != TrafficDirection.INCOMING:
            return 0

        if typepy.is_null_string(self.ifb_device):
            return -1

        return_code = 0
        modprobe_proc = spr.SubprocessRunner("modprobe ifb")

        try:
            if modprobe_proc.run() != 0:
                logger.error(modprobe_proc.stderr)
        except spr.CommandError as e:
            logger.debug(msgfy.to_debug_message(e))

        if self.is_add_shaping_rule or self.is_change_shaping_rule:
            notice_message = None
        else:
            notice_message = self.EXISTS_MSG_TEMPLATE.format(
                "failed to add ip link: ip link already exists."
            )

        return_code |= run_command_helper(
            "{:s} link add {:s} type ifb".format(find_bin_path("ip"), self.ifb_device),
            ignore_error_msg_regexp=self.REGEXP_FILE_EXISTS,
            notice_msg=notice_message,
        )

        return_code |= spr.SubprocessRunner(
            "{:s} link set dev {:s} up".format(find_bin_path("ip"), self.ifb_device)
        ).run()

        base_command = "{:s} add".format(get_tc_base_command(TcSubCommand.QDISC))
        if self.is_add_shaping_rule or self.is_change_shaping_rule:
            notice_message = None
        else:
            notice_message = self.EXISTS_MSG_TEMPLATE.format(
                "failed to '{:s}': ingress qdisc already exists.".format(base_command)
            )
        return_code |= run_command_helper(
            "{:s} dev {:s} ingress".format(base_command, self.device),
            ignore_error_msg_regexp=self.REGEXP_FILE_EXISTS,
            notice_msg=notice_message,
        )

        return_code |= spr.SubprocessRunner(
            " ".join(
                [
                    "{:s} add".format(get_tc_base_command(TcSubCommand.FILTER)),
                    "dev {:s}".format(self.device),
                    "parent ffff: protocol {:s} u32 match u32 0 0".format(self.protocol),
                    "flowid {:x}:".format(self.__qdisc_major_id),
                    "action mirred egress redirect",
                    "dev {:s}".format(self.ifb_device),
                ]
            )
        ).run()

        return return_code

    def __delete_qdisc(self):
        logging_msg = "delete {} qdisc".format(self.device)

        with logging_context(logging_msg):
            runner = spr.SubprocessRunner(
                "{:s} del dev {:s} root".format(
                    get_tc_base_command(TcSubCommand.QDISC), self.device
                )
            )
            runner.run()
            if re.search("RTNETLINK answers: No such file or directory", runner.stderr):
                logger.notice("no qdisc to delete for the outgoing device.")
                return False
            elif re.search("Cannot find device", runner.stderr):
                raise NetworkInterfaceNotFoundError(target=self.device)
            else:
                is_success = runner.returncode == 0
                if is_success:
                    logger.info(logging_msg)
                return is_success

    def __delete_ingress_qdisc(self):
        logging_msg = "delete {} ingress qdisc".format(self.device)

        with logging_context(logging_msg):
            returncode = run_command_helper(
                "{:s} del dev {:s} ingress".format(
                    get_tc_base_command(TcSubCommand.QDISC), self.device
                ),
                ignore_error_msg_regexp=re.compile(
                    "|".join(
                        [
                            "RTNETLINK answers: Invalid argument",  # debian/ubuntu
                            "RTNETLINK answers: No such file or directory",  # debian/ubuntu
                            "Error: Cannot find specified qdisc on specified device",  # RHEL/fedora
                        ]
                    )
                ),
                notice_msg="no qdisc to delete for the incoming device.",
            )
            is_success = returncode == 0
            if is_success:
                logger.info(logging_msg)

            return is_success

    def __delete_ifb_device(self):
        from ._capabilities import has_execution_authority, get_permission_error_message

        logging_msg = "delete {} ifb device ({})".format(self.device, self.ifb_device)
        with logging_context(logging_msg):
            verify_network_interface(self.ifb_device, self.__tc_command_output)

            if not has_execution_authority("ip"):
                logger.warn(get_permission_error_message("ip"))

                return errno.EPERM

            commands = [
                "{:s} del dev {:s} root".format(
                    get_tc_base_command(TcSubCommand.QDISC), self.ifb_device
                ),
                "{:s} link set dev {:s} down".format(find_bin_path("ip"), self.ifb_device),
                "{:s} link delete {:s} type ifb".format(find_bin_path("ip"), self.ifb_device),
            ]

            if all([spr.SubprocessRunner(command).run() != 0 for command in commands]):
                return 2

        logger.info(logging_msg)

        return 0


def delete_all_rules(device):
    TrafficControl(device).delete_all_rules()



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, division, print_function

import itertools

import pingparsing
import pytest
import typepy
from subprocrunner import SubprocessRunner

from tcconfig._const import Tc, TrafficDirection
from tcconfig.traffic_control import delete_all_rules

from .common import ASSERT_MARGIN, DEADLINE_TIME


@pytest.fixture
def device_option(request):
    return request.config.getoption("--device")


@pytest.fixture
def local_host_option(request):
    return request.config.getoption("--local-host")


@pytest.fixture
def dst_host_option(request):
    return request.config.getoption("--dst-host")


@pytest.fixture
def transmitter():
    transmitter = pingparsing.PingTransmitter()
    transmitter.ping_option = "-i 0.2 -q"
    transmitter.deadline = DEADLINE_TIME

    return transmitter


@pytest.fixture
def pingparser():
    return pingparsing.PingParsing()


class Test_tcset_one_network(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:

        python setup.py test --addopts "--device=<test device> \n
            --local-host <IP address> --dst-host=<IP address>"

    These tests expected to execute in the following environment:
       - Linux w/ iputils-ping package
       - English locale (for parsing ping output)
    """

    @pytest.mark.parametrize(
        ["shaping_algo", "delay"],
        [[params[0], params[1]] for params in itertools.product(["htb"], [100])],
    )
    def test_src_net_uniform_latency(
        self,
        device_option,
        local_host_option,
        dst_host_option,
        transmitter,
        pingparser,
        shaping_algo,
        delay,
    ):
        if device_option is None:
            pytest.skip("device option is null")
        if typepy.is_null_string(local_host_option):
            pytest.skip("local host is null")
        if typepy.is_null_string(dst_host_option):
            pytest.skip("destination host is null")

        for tc_target in [device_option, "--device {}".format(device_option)]:
            delete_all_rules(tc_target)
            transmitter.destination = dst_host_option

            # w/o latency tc ---
            without_tc_rtt_avg = pingparser.parse(transmitter.ping().stdout).rtt_avg

            # w/ latency tc ---
            assert (
                SubprocessRunner(
                    " ".join(
                        [
                            Tc.Command.TCSET,
                            tc_target,
                            "--src-network {:s}".format(local_host_option),
                            "--delay {:d}ms".format(delay),
                            "--shaping-algo {:s}".format(shaping_algo),
                        ]
                    )
                ).run()
                == 0
            )

            with_tc_rtt_avg = pingparser.parse(transmitter.ping().stdout).rtt_avg

            # assertion ---
            rtt_diff = with_tc_rtt_avg - without_tc_rtt_avg

            print("w/o tc rtt: {} ms".format(without_tc_rtt_avg))
            print("w/ tc rtt: {} ms".format(with_tc_rtt_avg))

            assert rtt_diff > (delay * ASSERT_MARGIN)

            # finalize ---
            delete_all_rules(tc_target)


class Test_tcset_exclude(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:

        python setup.py test --addopts "--device=<test device> \n
            --local-host <IP address> --dst-host=<IP address>"

    These tests expected to execute in the following environment:
       - Linux w/ iputils-ping package
       - English locale (for parsing ping output)
    """

    @pytest.mark.parametrize(
        ["shaping_algo", "delay"],
        [[params[0], params[1]] for params in itertools.product(["htb"], [100])],
    )
    def test_src_net_uniform_latency(
        self,
        device_option,
        local_host_option,
        dst_host_option,
        transmitter,
        pingparser,
        shaping_algo,
        delay,
    ):
        if device_option is None:
            pytest.skip("device option is null")
        if typepy.is_null_string(local_host_option):
            pytest.skip("local host is null")
        if typepy.is_null_string(dst_host_option):
            pytest.skip("destination host is null")

        for tc_target in [device_option, "--device {}".format(device_option)]:
            delete_all_rules(tc_target)
            transmitter.destination = dst_host_option

            # w/o latency tc ---
            without_tc_rtt_avg = pingparser.parse(transmitter.ping().stdout).rtt_avg

            # w/o latency tc (exclude network) ---
            assert (
                SubprocessRunner(
                    " ".join(
                        [
                            Tc.Command.TCSET,
                            tc_target,
                            "--direction {:s}".format(TrafficDirection.INCOMING),
                            "--exclude-dst-network {:s}".format(local_host_option),
                            "--exclude-src-network {:s}".format(dst_host_option),
                            "--delay {:d}ms".format(delay),
                            "--shaping-algo {:s}".format(shaping_algo),
                        ]
                    )
                ).run()
                == 0
            )

            exclude_tc_rtt_avg = pingparser.parse(transmitter.ping().stdout).rtt_avg

            # assertion ---
            rtt_diff = exclude_tc_rtt_avg - without_tc_rtt_avg

            print("w/o tc rtt: {} ms".format(without_tc_rtt_avg))
            print("exclude tc rtt: {} ms".format(exclude_tc_rtt_avg))

            assert rtt_diff < (delay / 10)

            # finalize ---
            delete_all_rules(tc_target)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import re

import pyparsing as pp
import simplejson as json
import typepy

from .._const import Tc, TcSubCommand
from .._logger import logger
from ._interface import AbstractParser


class TcQdiscParser(AbstractParser):
    __RE_DIRECT_QLEN = re.compile("direct_qlen (?P<number>[0-9]+)")

    @property
    def _tc_subcommand(self):
        return TcSubCommand.QDISC.value

    def __init__(self, con):
        super(TcQdiscParser, self).__init__()

        self.__con = con

    def __parse_direct_qlen(self, line):
        m = self.__RE_DIRECT_QLEN.search(line)
        if m is None:
            return

        self.__parsed_param["direct_qlen"] = int(m.group("number"))

    def parse(self, device, text):
        self._clear()

        if typepy.is_null_string(text):
            return []

        text = text.strip()
        entry_list = []

        for line in text.splitlines():
            if typepy.is_null_string(line):
                continue

            line = self._to_unicode(line.lstrip())

            if re.search("^qdisc netem |^qdisc htb |^qdisc tbf ", line) is None:
                continue

            if re.search("^qdisc htb ", line) is not None:
                self.__parse_direct_qlen(line)
                continue

            if re.search("^qdisc netem ", line) is not None:
                self.__parse_netem_param(line, "parent", pp.hexnums + ":")

            self.__parsed_param[Tc.Param.DEVICE] = device
            self.__parse_netem_param(line, "netem", pp.hexnums + ":", "handle")
            self.__parse_netem_param(line, "delay", pp.nums + ".msu")
            self.__parse_netem_delay_distro(line)
            self.__parse_netem_param(line, "loss", pp.nums + ".%")
            self.__parse_netem_param(line, "duplicate", pp.nums + ".%")
            self.__parse_netem_param(line, "corrupt", pp.nums + ".%")
            self.__parse_netem_param(line, "reorder", pp.nums + ".%")
            self.__parse_bandwidth_rate(line)

            logger.debug("parse a qdisc entry: {}".format(self.__parsed_param))

            entry_list.append(self.__parsed_param)

            self._clear()

        if entry_list:
            self.__con.create_table_from_data_matrix(
                self._tc_subcommand,
                [
                    Tc.Param.DEVICE,
                    "parent",
                    "handle",
                    "delay",
                    "delay-distro",
                    "loss",
                    "duplicate",
                    "corrupt",
                    "reorder",
                    "rate",
                ],
                entry_list,
            )

        logger.debug(
            "tc {:s} parse result: {}".format(self._tc_subcommand, json.dumps(entry_list, indent=4))
        )

        return entry_list

    def _clear(self):
        self.__parsed_param = {}

    def __parse_netem_delay_distro(self, line):
        parse_param_name = "delay"
        pattern = (
            pp.SkipTo(parse_param_name, include=True)
            + pp.Word(pp.nums + ".msu")
            + pp.Word(pp.nums + ".msu")
        )

        try:
            parsed_list = pattern.parseString(line)
            self.__parsed_param[parse_param_name] = parsed_list[2]
            self.__parsed_param["delay-distro"] = parsed_list[3]
        except pp.ParseException:
            pass

    def __parse_netem_param(self, line, parse_param_name, word_pattern, key_name=None):
        pattern = pp.SkipTo(parse_param_name, include=True) + pp.Word(word_pattern)
        if not key_name:
            key_name = parse_param_name

        try:
            result = pattern.parseString(line)[-1]
            if typepy.is_not_null_string(result):
                self.__parsed_param[key_name] = result
        except pp.ParseException:
            pass

    def __parse_bandwidth_rate(self, line):
        parse_param_name = "rate"
        pattern = pp.SkipTo(parse_param_name, include=True) + pp.Word(pp.alphanums + "." + ":")

        try:
            result = pattern.parseString(line)[-1]
            if typepy.is_not_null_string(result):
                result = result.rstrip("bit")
                self.__parsed_param[parse_param_name] = result
        except pp.ParseException:
            pass



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import

import abc

from pyroute2 import IPRoute


class TargetNotFoundError(Exception):
    @abc.abstractproperty
    def _target_type(self):
        return None

    def __init__(self, *args, **kwargs):
        self._target = kwargs.pop("target", None)

        super(TargetNotFoundError, self).__init__(*args, **kwargs)

    def __str__(self, *args, **kwargs):
        item_list = [Exception.__str__(self, *args, **kwargs)]

        if self._target:
            item_list.append("{} not found: {}".format(self._target_type, self._target))

        return " ".join(item_list).strip()

    def __repr__(self, *args, **kwargs):
        return self.__str__(*args, **kwargs)


class NetworkInterfaceNotFoundError(TargetNotFoundError):
    """
    Exception raised when network interface not found.
    """

    @property
    def _target_type(self):
        return "network interface"

    def __str__(self, *args, **kwargs):
        item_list = [super(NetworkInterfaceNotFoundError, self).__str__(*args, **kwargs)]

        with IPRoute() as ipr:
            avail_interfaces = [link.get_attr("IFLA_IFNAME") for link in ipr.get_links()]

        item_list.append("(available interfaces: {})".format(", ".join(avail_interfaces)))

        return " ".join(item_list).strip()


class ContainerNotFoundError(TargetNotFoundError):
    """
    Exception raised when container not found.
    """

    @property
    def _target_type(self):
        return "container"

    def __str__(self, *args, **kwargs):
        from ._docker import DockerClient

        dclient = DockerClient()
        container_names = dclient.extract_running_container_names()
        item_list = [super(ContainerNotFoundError, self).__str__(*args, **kwargs)]

        if container_names:
            item_list.append(
                "(available running containers: {})".format(", ".join(container_names))
            )
        else:
            item_list.append("(running container not found)")

        return " ".join(item_list).strip()


class ModuleNotFoundError(Exception):
    """
    Exception raised when mandatory kernel module not found.
    """


class TcCommandExecutionError(Exception):
    """
    Exception raised when failed to execute a ``tc`` command.
    """


class TcAlreadyExist(TcCommandExecutionError):
    """
    Exception raised when a traffic shaping rule already exist.
    """



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class ParserInterface(object):
    @abc.abstractmethod
    def parse(self, device, text):  # pragma: no cover
        pass


@six.add_metaclass(abc.ABCMeta)
class AbstractParser(ParserInterface):
    def __init__(self):
        self._clear()

    @abc.abstractproperty
    def _tc_subcommand(self):  # pragma: no cover
        pass

    @abc.abstractmethod
    def _clear(self):  # pragma: no cover
        pass

    @staticmethod
    def _to_unicode(text):
        try:
            return text.decode("ascii")
        except AttributeError:
            return text



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import subprocrunner as spr

from ._common import find_bin_path
from ._const import TcSubCommand
from ._error import NetworkInterfaceNotFoundError


def get_tc_base_command(tc_subcommand):
    if not isinstance(tc_subcommand, TcSubCommand):
        raise ValueError("the argument must be a TcSubCommand value")

    return "{:s} {:s}".format(find_bin_path("tc"), tc_subcommand.value)


def run_tc_show(subcommand, device, tc_command_output):
    from ._network import verify_network_interface

    verify_network_interface(device, tc_command_output)

    runner = spr.SubprocessRunner(
        "{:s} show dev {:s}".format(get_tc_base_command(subcommand), device)
    )
    if runner.run() != 0 and runner.stderr.find("Cannot find device") != -1:
        # reach here if the device does not exist at the system and netiface
        # not installed.
        raise NetworkInterfaceNotFoundError(target=device)

    return runner.stdout



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import contextlib
import errno
import os
import re
import sys

import logbook
import logbook.more
import msgfy
import subprocrunner as spr
import typepy
from humanreadable import ParameterError
from path import Path
from simplesqlite import SimpleSQLite

from ._const import IPV6_OPTION_ERROR_MSG_FORMAT, TcCommandOutput
from ._logger import logger


_bin_path_cache = {}


@contextlib.contextmanager
def logging_context(name):
    logger.debug("|---- {:s}: {:s} -----".format("start", name))
    try:
        yield
    finally:
        logger.debug("----- {:s}: {:s} ----|".format("complete", name))


def find_bin_path(command):
    def _to_regular_bin_path(file_path):
        path_obj = Path(file_path)
        if path_obj.islink():
            return path_obj.readlinkabs()

        return file_path

    if command in _bin_path_cache:
        return _bin_path_cache.get(command)

    bin_path = spr.Which(command)
    if bin_path.is_exist():
        _bin_path_cache[command] = _to_regular_bin_path(bin_path.abspath())
        return _bin_path_cache[command]

    for sbin_path in ("/sbin/{:s}".format(command), "/usr/sbin/{:s}".format(command)):
        if os.path.isfile(sbin_path):
            _bin_path_cache[command] = _to_regular_bin_path(sbin_path)
            return _bin_path_cache[command]

    # return the command as it is when binary file not found
    return command


def check_command_installation(command):
    if find_bin_path(command):
        return

    logger.error("command not found: {}".format(command))
    sys.exit(errno.ENOENT)


def initialize_cli(options):
    from ._logger import set_log_level

    debug_level_format_str = (
        "[{record.level_name}] {record.channel} {record.func_name} "
        "({record.lineno}): {record.message}"
    )
    if options.log_level == logbook.DEBUG:
        info_level_format_str = debug_level_format_str
    else:
        info_level_format_str = "[{record.level_name}] {record.channel}: {record.message}"

    logbook.more.ColorizedStderrHandler(
        level=logbook.DEBUG, format_string=debug_level_format_str
    ).push_application()
    logbook.more.ColorizedStderrHandler(
        level=logbook.INFO, format_string=info_level_format_str
    ).push_application()

    set_log_level(options.log_level)
    spr.SubprocessRunner.is_save_history = True

    if options.is_output_stacktrace:
        spr.SubprocessRunner.is_output_stacktrace = options.is_output_stacktrace

    SimpleSQLite.global_debug_query = options.debug_query


def is_execute_tc_command(tc_command_output):
    return tc_command_output == TcCommandOutput.NOT_SET


def validate_within_min_max(param_name, value, min_value, max_value, unit):
    from dataproperty import DataProperty

    if value is None:
        return

    if unit is None:
        unit = ""
    else:
        unit = "[{:s}]".format(unit)

    if value > max_value:
        raise ParameterError(
            "'{:s}' is too high".format(param_name),
            expected="<={:s}{:s}".format(DataProperty(max_value).to_str(), unit),
            value="{:s}{:s}".format(DataProperty(value).to_str(), unit),
        )

    if value < min_value:
        raise ParameterError(
            "'{:s}' is too low".format(param_name),
            expected=">={:s}{:s}".format(DataProperty(min_value).to_str(), unit),
            value="{:s}{:s}".format(DataProperty(value).to_str(), unit),
        )


def normalize_tc_value(tc_obj):
    import ipaddress

    try:
        tc_obj.sanitize()
    except ipaddress.AddressValueError as e:
        logger.error(IPV6_OPTION_ERROR_MSG_FORMAT.format(e))
        sys.exit(errno.EINVAL)
    except ValueError as e:
        logger.error(msgfy.to_error_message(e))
        sys.exit(errno.EINVAL)


def run_command_helper(command, ignore_error_msg_regexp, notice_msg, exception_class=None):
    proc = spr.SubprocessRunner(command, error_log_level=logbook.NOTSET)
    proc.run()

    if proc.returncode == 0:
        return 0

    if ignore_error_msg_regexp:
        match = ignore_error_msg_regexp.search(proc.stderr)
        if match is None:
            error_msg = "\n".join(
                [
                    "command execution failed",
                    "  command={}".format(command),
                    "  stderr={}".format(proc.stderr),
                ]
            )

            if re.search("RTNETLINK answers: Operation not permitted", proc.stderr):
                logger.error(error_msg)
                sys.exit(proc.returncode)

            logger.error(error_msg)

            return proc.returncode

    if typepy.is_not_null_string(notice_msg):
        logger.notice(notice_msg)

    if exception_class is not None:
        raise exception_class(command)

    return proc.returncode



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, print_function

import json
import sys
from collections import OrderedDict

import humanreadable as hr
from subprocrunner import SubprocessRunner
from typepy import RealNumber


DEADLINE_TIME = 3  # [sec]
ASSERT_MARGIN = 0.5


def print_test_result(expected, actual, error=None):
    if isinstance(expected, (dict, OrderedDict)):
        expected = json.dumps(expected, indent=4)
    print("[expected]\n{}\n".format(expected))

    if isinstance(actual, (dict, OrderedDict)):
        actual = json.dumps(actual, indent=4)
    print("[actual]\n{}\n".format(actual))

    if error:
        print(error, file=sys.stderr)


def is_invalid_param(rate, delay, packet_loss, packet_duplicate, corrupt, reordering):
    param_values = [packet_loss, packet_duplicate, corrupt, reordering]

    print("rate={}, params={}".format(rate, param_values))

    is_invalid = all(
        [not RealNumber(param_value).is_type() or param_value <= 0 for param_value in param_values]
        + [hr.Time(delay, hr.Time.Unit.MILLISECOND).milliseconds <= 0]
    )

    try:
        hr.BitPerSecond(rate).bps
    except (TypeError, ValueError):
        pass
    else:
        is_invalid = False

    return is_invalid


def runner_helper(command):
    proc_runner = SubprocessRunner(command)
    proc_runner.run()

    print("{}\n{}\n".format(proc_runner.command_str, proc_runner.stderr), file=sys.stderr)

    assert proc_runner.returncode == 0

    return proc_runner


class NullLogger(object):
    level_name = None

    def catch_exceptions(self, *args, **kwargs):  # pragma: no cover
        pass

    def critical(self, *args, **kwargs):  # pragma: no cover
        pass

    def debug(self, *args, **kwargs):  # pragma: no cover
        pass

    def disable(self):  # pragma: no cover
        pass

    def enable(self):  # pragma: no cover
        pass

    def error(self, *args, **kwargs):  # pragma: no cover
        pass

    def exception(self, *args, **kwargs):  # pragma: no cover
        pass

    def info(self, *args, **kwargs):  # pragma: no cover
        pass

    def log(self, level, *args, **kwargs):  # pragma: no cover
        pass

    def notice(self, *args, **kwargs):  # pragma: no cover
        pass

    def warn(self, *args, **kwargs):  # pragma: no cover
        pass

    def warning(self, *args, **kwargs):  # pragma: no cover
        pass



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, print_function, unicode_literals

import sys

import msgfy
import simplejson as json
import subprocrunner as spr
from docker.errors import DockerException
from simplesqlite import SimpleSQLite
from simplesqlite.model import Integer, Model, Text

from .__version__ import __version__
from ._argparse_wrapper import ArgparseWrapper
from ._common import check_command_installation, initialize_cli
from ._const import Tc, TcCommandOutput
from ._docker import DockerClient
from ._error import TargetNotFoundError
from ._logger import logger
from ._network import verify_network_interface
from ._tc_script import write_tc_script
from .parser.shaping_rule import TcShapingRuleParser


class ShapingRuleModel(Model):
    device = Text(not_null=True)
    direction = Text(not_null=True)
    filter_id = Text(not_null=True)
    dst_network = Text()
    dst_port = Integer()
    src_network = Text()
    src_port = Integer()
    protocol = Text(not_null=True)

    delay = Text()
    delay_distro = Text()
    loss = Text()
    duplicate = Text()
    corrupt = Text()
    reorder = Text()
    rate = Text()


def parse_option():
    parser = ArgparseWrapper(__version__)

    group = parser.parser.add_argument_group("Traffic Control")
    if set(["-d", "--device"]).intersection(set(sys.argv)):
        # deprecated: remain for backward compatibility
        group.add_argument(
            "-d", "--device", action="append", required=True, help="network device name (e.g. eth0)"
        )
    else:
        group.add_argument("device", nargs="+", help="network device name (e.g. eth0)")
    group.add_argument(
        "--ipv6",
        dest="ip_version",
        action="store_const",
        const=6,
        default=4,
        help="Display IPv6 shaping rules. Defaults to show IPv4 shaping rules.",
    )

    parser.add_docker_group(is_add_srcdst=False)

    parser.parser.add_argument(
        "--color",
        action="store_true",
        default=False,
        help="colorize the output. require Pygments package.",
    )
    parser.parser.add_argument("--export", dest="export_path", help="[experimental]")
    parser.parser.add_argument(
        "--exclude-filter-id",
        action="store_true",
        default=False,
        help="[experimental] not display filter_id.",
    )

    return parser.parser.parse_args()


def print_tc(text, is_colorize):
    try:
        from pygments import highlight
        from pygments.lexers import JsonLexer
        from pygments.formatters import TerminalTrueColorFormatter

        pygments_installed = True
    except ImportError:
        pygments_installed = False

    if is_colorize and pygments_installed:
        print(
            highlight(
                code=text, lexer=JsonLexer(), formatter=TerminalTrueColorFormatter(style="monokai")
            )
        )
    else:
        print(text)


def export_settings(export_path, out_rules, in_rules):
    with SimpleSQLite(export_path, mode="a") as con:
        ShapingRuleModel.attach(con)
        ShapingRuleModel.create()

        for out_rule in out_rules:
            ShapingRuleModel.insert(ShapingRuleModel(**out_rule))

        for in_rule in in_rules:
            ShapingRuleModel.insert(ShapingRuleModel(**in_rule))


def extract_tc_params(options):
    dclient = None
    if options.use_docker:
        try:
            dclient = DockerClient(options.tc_command_output)
        except DockerException as e:
            logger.error(msgfy.to_error_message(e))
            sys.exit(1)

    tc_params = {}

    for device in options.device:
        try:
            if options.use_docker:
                container = device
                container_info = dclient.extract_container_info(container)

                if not container_info.state.running:
                    logger.error(
                        "{id} ({name}) not running (current status: {status})".format(
                            id=container,
                            name=container_info.name,
                            status=container_info.state.status,
                        )
                    )
                    continue

                dclient.create_veth_table(container)

                for veth in dclient.fetch_veth_list(container_info.name):
                    rule_parser = TcShapingRuleParser(
                        device=veth,
                        ip_version=options.ip_version,
                        tc_command_output=options.tc_command_output,
                        logger=logger,
                        export_path=options.export_path,
                        is_parse_filter_id=not options.exclude_filter_id,
                    )
                    rule_parser.parse()

                    if options.export_path:
                        rule_parser.con.dump(options.export_path)
                        out_rules, in_rules = rule_parser.extract_export_parameters()
                        export_settings(options.export_path, out_rules, in_rules)

                    tc_params.update(rule_parser.get_tc_parameter())
                    key = "{id} (device={veth})".format(id=container_info.id[:12], veth=veth,)
                    tc_params[key] = tc_params.pop(veth)
            else:
                verify_network_interface(device, options.tc_command_output)
                rule_parser = TcShapingRuleParser(
                    device=device,
                    ip_version=options.ip_version,
                    tc_command_output=options.tc_command_output,
                    logger=logger,
                    export_path=options.export_path,
                    is_parse_filter_id=not options.exclude_filter_id,
                )
                rule_parser.parse()

                if options.export_path:
                    rule_parser.con.dump(options.export_path)
                    out_rules, in_rules = rule_parser.extract_export_parameters()
                    export_settings(options.export_path, out_rules, in_rules)

                tc_params.update(rule_parser.get_tc_parameter())
        except TargetNotFoundError as e:
            logger.warn(e)
            continue

    return tc_params


def main():
    options = parse_option()

    initialize_cli(options)
    check_command_installation("tc")

    if options.tc_command_output != TcCommandOutput.NOT_SET:
        spr.SubprocessRunner.default_is_dry_run = True

    tc_params = extract_tc_params(options)
    command_history = "\n".join(spr.SubprocessRunner.get_history())

    if options.tc_command_output == TcCommandOutput.STDOUT:
        print(command_history)
        return 0

    if options.tc_command_output == TcCommandOutput.SCRIPT:
        write_tc_script(
            Tc.Command.TCSHOW, command_history, filename_suffix="-".join(options.device)
        )
        return 0

    logger.debug("command history\n{}".format(command_history))

    print_tc(json.dumps(tc_params, ensure_ascii=False, indent=4), options.color)

    return 0


if __name__ == "__main__":
    sys.exit(main())



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import errno
import io
import re

import msgfy
import pyparsing as pp
import simplejson as json
import six
import subprocrunner as spr

from ._const import Network, Tc, TrafficDirection


RE_CONTAINER_ID = re.compile(r"[a-z0-9]{12}\s+\(device=[a-z0-9]+\)")
# e.g. edfd9dbb3969 (device=veth6f7b798)


class TcConfigLoader(object):
    def __init__(self, logger):
        self.__logger = logger
        self.__config_table = None
        self.is_overwrite = False

    def load_tcconfig(self, config_file_path):
        from voluptuous import Schema, Required, Any, ALLOW_EXTRA

        schema = Schema(
            {
                Required(six.text_type): {
                    Any(*TrafficDirection.LIST): {
                        six.text_type: {six.text_type: Any(six.text_type, int, float)}
                    }
                }
            },
            extra=ALLOW_EXTRA,
        )

        with io.open(config_file_path, encoding="utf-8") as fp:
            self.__config_table = json.load(fp)

        schema(self.__config_table)
        self.__logger.debug(
            "tc config file: {:s}".format(json.dumps(self.__config_table, indent=4))
        )

    def get_tcconfig_commands(self):
        from .tcset import get_arg_parser

        command_list = []

        for device, device_table in six.iteritems(self.__config_table):
            is_container = RE_CONTAINER_ID.search(device) is not None
            if is_container:
                device = device.split()[0]

            if self.is_overwrite:
                command_list.append(
                    " ".join(
                        [Tc.Command.TCDEL, device, "--all"] + (["--docker"] if is_container else [])
                    )
                )

            for direction, direction_table in six.iteritems(device_table):
                is_first_set = True

                for tc_filter, filter_table in six.iteritems(direction_table):
                    self.__logger.debug(
                        "is_first_set={}, filter='{}', table={}".format(
                            is_first_set, tc_filter, filter_table
                        )
                    )

                    if not filter_table:
                        continue

                    option_list = [device, "--direction={:s}".format(direction)] + (
                        ["--docker"] if is_container else []
                    )

                    for key, value in six.iteritems(filter_table):
                        arg_item = "--{:s}={}".format(key, value)

                        parse_result = get_arg_parser().parse_known_args(["dummy", arg_item])
                        if parse_result[1]:
                            self.__logger.debug(
                                "unknown parameter: key={}, value={}".format(key, value)
                            )
                            continue

                        option_list.append(arg_item)

                    try:
                        dst_network = self.__parse_tc_filter_network(tc_filter)
                        if dst_network not in (Network.Ipv4.ANYWHERE, Network.Ipv6.ANYWHERE):
                            option_list.append("--dst-network={:s}".format(dst_network))
                    except pp.ParseException:
                        pass

                    try:
                        src_port = self.__parse_tc_filter_src_port(tc_filter)
                        option_list.append("--src-port={}".format(src_port))
                    except pp.ParseException:
                        pass

                    try:
                        dst_port = self.__parse_tc_filter_dst_port(tc_filter)
                        option_list.append("--dst-port={}".format(dst_port))
                    except pp.ParseException:
                        pass

                    if not is_first_set:
                        option_list.append("--add")

                    is_first_set = False

                    command_list.append(" ".join([Tc.Command.TCSET] + option_list))

        return command_list

    @staticmethod
    def __parse_tc_filter_network(text):
        network_pattern = pp.SkipTo("{:s}=".format(Tc.Param.DST_NETWORK), include=True) + pp.Word(
            pp.alphanums + "." + "/"
        )

        return network_pattern.parseString(text)[-1]

    @staticmethod
    def __parse_tc_filter_src_port(text):
        port_pattern = pp.SkipTo("{:s}=".format(Tc.Param.SRC_PORT), include=True) + pp.Word(pp.nums)

        return port_pattern.parseString(text)[-1]

    @staticmethod
    def __parse_tc_filter_dst_port(text):
        port_pattern = pp.SkipTo("{:s}=".format(Tc.Param.DST_PORT), include=True) + pp.Word(pp.nums)

        return port_pattern.parseString(text)[-1]


def set_tc_from_file(logger, config_file_path, is_overwrite):
    return_code = 0

    loader = TcConfigLoader(logger)
    loader.is_overwrite = is_overwrite

    try:
        loader.load_tcconfig(config_file_path)
    except IOError as e:
        logger.error(msgfy.to_error_message(e))
        return errno.EIO

    for tcconfig_command in loader.get_tcconfig_commands():
        return_code |= spr.SubprocessRunner(tcconfig_command).run()

    return return_code



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, print_function, unicode_literals

import errno
import ipaddress
import sys

import humanreadable as hr
import logbook
import msgfy
import subprocrunner as spr

from .__version__ import __version__
from ._argparse_wrapper import ArgparseWrapper
from ._capabilities import check_execution_authority
from ._common import initialize_cli, is_execute_tc_command, normalize_tc_value
from ._const import IPV6_OPTION_ERROR_MSG_FORMAT, ShapingAlgorithm, Tc, TrafficDirection
from ._error import ContainerNotFoundError, ModuleNotFoundError, NetworkInterfaceNotFoundError
from ._importer import set_tc_from_file
from ._logger import logger, set_log_level
from ._main import Main
from ._netem_param import (
    MAX_CORRUPTION_RATE,
    MAX_PACKET_DUPLICATE_RATE,
    MAX_PACKET_LOSS_RATE,
    MAX_REORDERING_RATE,
    MIN_CORRUPTION_RATE,
    MIN_PACKET_DUPLICATE_RATE,
    MIN_PACKET_LOSS_RATE,
    MIN_REORDERING_RATE,
    NetemParameter,
)
from ._shaping_rule_finder import TcShapingRuleFinder
from .traffic_control import TrafficControl


def _get_unit_help_msg():
    return ", ".join(["/".join(values) for values in hr.Time.get_text_units().values()])


def get_arg_parser():
    parser = ArgparseWrapper(__version__)

    if set(["-d", "--device"]).intersection(set(sys.argv)):
        # deprecated: remain for backward compatibility
        parser.parser.add_argument(
            "-d", "--device", required=True, help="network device name (e.g. eth0)"
        )
    else:
        parser.parser.add_argument(
            "device", help="target name: network-interface/config-file (e.g. eth0)"
        )

    parser.parser.add_argument(
        "--import-setting",
        action="store_true",
        default=False,
        help="import traffic control settings from a configuration file.",
    )

    group = parser.parser.add_mutually_exclusive_group()
    group.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="overwrite existing traffic shaping rules.",
    )
    group.add_argument(
        "--change",
        dest="is_change_shaping_rule",
        action="store_true",
        default=False,
        help="""change existing traffic shaping rules to the new one.
        this option is effective to reduce the time between the shaping rule switching
        compared to --overwrite option.
        note: just adds a shaping rule if there are no existing shaping rules.
        """,
    )
    group.add_argument(
        "--add",
        dest="is_add_shaping_rule",
        action="store_true",
        default=False,
        help="add a traffic shaping rule in addition to existing rules.",
    )

    group = parser.parser.add_argument_group("Traffic Control Parameters")
    group.add_argument(
        "--rate",
        "--bandwidth-rate",
        dest="bandwidth_rate",
        help="""network bandwidth rate [bit per second].
        the minimum bandwidth rate is 8 bps.
        valid units are either: {}.
        e.g. tcset eth0 --rate 10Mbps
        """.format(
            ", ".join([", ".join(values) for values in hr.BitPerSecond.get_text_units().values()])
        ),
    )
    group.add_argument(
        "--delay",
        dest="network_latency",
        default=Tc.ValueRange.LatencyTime.MIN,
        help="""round trip network delay. the valid range is from {min_value:} to {max_value:}.
        valid time units are: {unit}. if no unit string found, considered milliseconds as
        the time unit. (default=%(default)s)
        """.format(
            min_value=Tc.ValueRange.LatencyTime.MIN,
            max_value=Tc.ValueRange.LatencyTime.MAX,
            unit=_get_unit_help_msg(),
        ),
    )
    group.add_argument(
        "--delay-distro",
        dest="latency_distro_time",
        default=Tc.ValueRange.LatencyTime.MIN,
        help="""distribution of network latency becomes X +- Y (normal distribution).
        Here X is the value of --delay option and Y is the value of --delay-dist option).
        network latency distribution is uniform, without this option. valid time units are: {unit}.
        if no unit string found, considered milliseconds as the time unit.
        """.format(
            unit=_get_unit_help_msg()
        ),
    )
    group.add_argument(
        "--loss",
        dest="packet_loss_rate",
        default=0,
        help="""round trip packet loss rate [%%]. the valid range is from {:d}
        to {:d}. (default=%(default)s)
        """.format(
            MIN_PACKET_LOSS_RATE, MAX_PACKET_LOSS_RATE
        ),
    )
    group.add_argument(
        "--duplicate",
        dest="packet_duplicate_rate",
        default=0,
        help="""round trip packet duplicate rate [%%]. the valid range is
        from {:d} to {:d}. (default=%(default)s)
        """.format(
            MIN_PACKET_DUPLICATE_RATE, MAX_PACKET_DUPLICATE_RATE
        ),
    )
    group.add_argument(
        "--corrupt",
        dest="corruption_rate",
        default=0,
        help="""packet corruption rate [%%]. the valid range is from {:d} to {:d}.
        packet corruption means single bit error at a random offset in the packet.
        (default=%(default)s)
        """.format(
            MIN_CORRUPTION_RATE, MAX_CORRUPTION_RATE
        ),
    )
    group.add_argument(
        "--reordering",
        dest="reordering_rate",
        default=0,
        help="""packet reordering rate [%%]. the valid range is from {:d}
        to {:d}. (default=%(default)s)
        """.format(
            MIN_REORDERING_RATE, MAX_REORDERING_RATE
        ),
    )
    group.add_argument(
        "--shaping-algo",
        dest="shaping_algorithm",
        choices=[ShapingAlgorithm.HTB, ShapingAlgorithm.TBF],
        default=ShapingAlgorithm.HTB,
        help="shaping algorithm. defaults to %(default)s (recommended).",
    )
    group.add_argument(
        "--iptables",
        dest="is_enable_iptables",
        action="store_true",
        default=False,
        help="use iptables to traffic control.",
    )

    group = parser.add_routing_group()
    group.add_argument(
        "--exclude-dst-network",
        help="exclude a specific destination IP-address/network from a shaping rule.",
    )
    group.add_argument(
        "--exclude-src-network",
        help="exclude a specific source IP-address/network from a shaping rule.",
    )
    group.add_argument(
        "--exclude-dst-port", help="exclude a specific destination port from a shaping rule."
    )
    group.add_argument(
        "--exclude-src-port", help="exclude a specific source port from a shaping rule."
    )

    parser.add_docker_group()

    return parser.parser


def verify_netem_module():
    import re

    runner = spr.SubprocessRunner("lsmod")

    try:
        if runner.run() != 0:
            raise OSError(runner.returncode, "failed to execute lsmod")
    except spr.CommandError as e:
        # reach here when the kmod package not installed.
        # this kind of environments could exist such as slim Docker containers.
        logger.debug(msgfy.to_debug_message(e))
    else:
        if re.search(r"\bsch_netem\b", runner.stdout) is None:
            raise ModuleNotFoundError("sch_netem module not found")


class TcSetMain(Main):
    def run(self):
        return_code_list = []

        for device in self._fetch_tc_targets():
            tc = self.__create_tc(device)
            return_code = self.__check_tc(tc)

            if return_code != 0:
                return_code_list.append(return_code)
                continue

            normalize_tc_value(tc)

            if self._options.overwrite:
                if self._options.log_level == logbook.INFO:
                    set_log_level(logbook.ERROR)

                try:
                    tc.delete_all_rules()
                except NetworkInterfaceNotFoundError:
                    pass

                set_log_level(self._options.log_level)

            if (
                self._options.is_add_shaping_rule
                and TcShapingRuleFinder(logger=logger, tc=tc).is_exist_rule()
            ):
                logger.error(
                    "\n".join(
                        [
                            "adding a shaping rule failed. a shaping rule for the same "
                            "network/port already exist. try to execute with:",
                            "  (a) --overwrite option if you want to overwrite "
                            "the existing rules.",
                            "  (b) --change option if you want to change "
                            "the existing rule parameters.",
                        ]
                    )
                )
                return errno.EINVAL

            try:
                return_code_list.append(tc.set_shaping_rule())
            except NetworkInterfaceNotFoundError as e:
                logger.error(e)
                return errno.EINVAL

            self._dump_history(tc, Tc.Command.TCSET)

        return self._get_return_code(return_code_list)

    def __check_tc(self, tc):
        try:
            tc.validate()
        except (NetworkInterfaceNotFoundError, ContainerNotFoundError) as e:
            logger.error(e)
            return errno.EINVAL
        except ipaddress.AddressValueError as e:
            logger.error(IPV6_OPTION_ERROR_MSG_FORMAT.format(e))
            return errno.EINVAL
        except hr.ParameterError as e:
            logger.error(msgfy.to_error_message(e))
            return errno.EINVAL

        return 0

    def __create_tc(self, device):
        options = self._options

        return TrafficControl(
            device,
            direction=options.direction,
            netem_param=NetemParameter(
                device=device,
                bandwidth_rate=options.bandwidth_rate,
                latency_time=options.network_latency,
                latency_distro_time=options.latency_distro_time,
                packet_loss_rate=options.packet_loss_rate,
                packet_duplicate_rate=options.packet_duplicate_rate,
                corruption_rate=options.corruption_rate,
                reordering_rate=options.reordering_rate,
            ),
            dst_network=self._extract_dst_network(),
            exclude_dst_network=options.exclude_dst_network,
            src_network=self._extract_src_network(),
            exclude_src_network=options.exclude_src_network,
            src_port=options.src_port,
            exclude_src_port=options.exclude_src_port,
            dst_port=options.dst_port,
            exclude_dst_port=options.exclude_dst_port,
            is_ipv6=options.is_ipv6,
            is_change_shaping_rule=options.is_change_shaping_rule,
            is_add_shaping_rule=options.is_add_shaping_rule,
            is_enable_iptables=options.is_enable_iptables,
            shaping_algorithm=options.shaping_algorithm,
            tc_command_output=options.tc_command_output,
        )


def main():
    options = get_arg_parser().parse_args()

    initialize_cli(options)

    if is_execute_tc_command(options.tc_command_output):
        check_execution_authority("tc")

        if options.direction == TrafficDirection.INCOMING:
            check_execution_authority("ip")
    else:
        spr.SubprocessRunner.default_is_dry_run = True

    try:
        verify_netem_module()
    except ModuleNotFoundError as e:
        logger.debug(e)

    if options.import_setting:
        return set_tc_from_file(logger, options.device, options.overwrite)

    spr.SubprocessRunner.clear_history()

    return TcSetMain(options).run()


if __name__ == "__main__":
    sys.exit(main())



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import ipaddress
import re
from collections import OrderedDict

import pyparsing as pp
import simplejson as json
import typepy
from six import text_type

from .._const import Tc, TcSubCommand
from .._logger import logger
from .._network import sanitize_network
from ._interface import AbstractParser


class TcFilterParser(AbstractParser):
    class FilterMatchIdIpv4(object):
        INCOMING_NETWORK = 12
        OUTGOING_NETWORK = 16
        PORT = 20

    class FilterMatchIdIpv6(object):
        INCOMING_NETWORK_LIST = [8, 12, 16, 20]
        OUTGOING_NETWORK_LIST = [24, 28, 32, 36]
        PORT = 40

    __FILTER_FLOWID_PATTERN = (
        pp.Literal("filter parent") + pp.SkipTo("flowid", include=True) + pp.Word(pp.hexnums + ":")
    )
    __FILTER_PROTOCOL_PATTERN = (
        pp.Literal("filter parent") + pp.SkipTo("protocol", include=True) + pp.Word(pp.alphanums)
    )
    __FILTER_PRIORITY_PATTERN = (
        pp.Literal("filter parent") + pp.SkipTo("pref", include=True) + pp.Word(pp.nums)
    )
    __FILTER_ID_PATTERN = (
        pp.Literal("filter parent") + pp.SkipTo("fh", include=True) + pp.Word(pp.hexnums + ":")
    )
    __FILTER_MATCH_PATTERN = (
        pp.Literal("match") + pp.Word(pp.alphanums + "/") + pp.Literal("at") + pp.Word(pp.nums)
    )
    __FILTER_MANGLE_MARK_PATTERN = (
        pp.Literal("filter parent")
        + pp.SkipTo("handle", include=True)
        + pp.Word(pp.hexnums)
        + pp.SkipTo("classid", include=True)
        + pp.Word(pp.hexnums + ":")
    )

    @property
    def protocol(self):
        return self.__protocol

    @property
    def _tc_subcommand(self):
        return TcSubCommand.FILTER.value

    def __init__(self, con, ip_version):
        super(TcFilterParser, self).__init__()

        self.__con = con
        self.__ip_version = ip_version
        self.__buffer = None
        self.__parse_idx = 0

        self.__protocol = None

        self._clear()

    def parse(self, device, text):
        self._clear()

        if typepy.is_null_string(text):
            return []

        filter_data_matrix = []
        self.__buffer = self._to_unicode(text).splitlines()
        self.__parse_idx = 0

        while self.__parse_idx < len(self.__buffer):
            line = self._to_unicode(self.__buffer[self.__parse_idx].strip())
            self.__parse_idx += 1

            if typepy.is_null_string(line):
                continue

            self.__device = device

            try:
                self.__parse_mangle_mark(line)
            except pp.ParseException:
                logger.debug("failed to parse mangle: {}".format(line))
            else:
                filter_data_matrix.append(
                    {
                        Tc.Param.DEVICE: self.__device,
                        Tc.Param.CLASS_ID: self.__classid,
                        Tc.Param.HANDLE: self.__handle,
                    }
                )
                self._clear()
                continue

            tc_filter = self.__get_filter()

            try:
                self.__parse_flow_id(line)
                self.__parse_protocol(line)
                self.__parse_priority(line)
                self.__parse_filter_id(line)

                if tc_filter.get(Tc.Param.FLOW_ID):
                    logger.debug("store filter: {}".format(tc_filter))
                    filter_data_matrix.append(tc_filter)
                    self._clear()

                    self.__parse_flow_id(line)
                    self.__parse_protocol(line)
                    self.__parse_priority(line)
                    self.__parse_filter_id(line)

                continue
            except pp.ParseException:
                logger.debug("failed to parse flow id: {}".format(line))

            try:
                if self.__ip_version == 4:
                    self.__parse_filter_ipv4(line)
                elif self.__ip_version == 6:
                    self.__parse_filter_ipv6(line)
                else:
                    raise ValueError("unknown ip version: {}".format(self.__ip_version))
            except pp.ParseException:
                logger.debug("failed to parse filter: {}".format(line))

        if self.__flow_id:
            filter_data_matrix.append(self.__get_filter())

        if filter_data_matrix:
            self.__con.create_table_from_data_matrix(
                self._tc_subcommand, list(self.__get_filter()), filter_data_matrix
            )

        logger.debug(
            "tc {:s} parse result: {}".format(
                self._tc_subcommand, json.dumps(filter_data_matrix, indent=4)
            )
        )

        return filter_data_matrix

    def parse_incoming_device(self, text):
        if typepy.is_null_string(text):
            return None

        match = re.search(
            r"Egress Redirect to device ifb[\d]+", self._to_unicode(text), re.MULTILINE
        )
        if match is None:
            return None

        return re.search(r"ifb[\d]+", match.group()).group()

    def _clear(self):
        self.__device = None
        self.__filter_id = None
        self.__flow_id = None
        self.__protocol = None
        self.__priority = None
        self.__filter_src_network = None
        self.__filter_dst_network = None
        self.__filter_src_port = None
        self.__filter_dst_port = None

        self.__handle = None
        self.__classid = None

    def __get_filter(self):
        tc_filter = OrderedDict()
        tc_filter[Tc.Param.DEVICE] = self.__device
        tc_filter[Tc.Param.FILTER_ID] = self.__filter_id
        tc_filter[Tc.Param.FLOW_ID] = self.__flow_id
        tc_filter[Tc.Param.PROTOCOL] = self.protocol
        tc_filter[Tc.Param.PRIORITY] = self.__priority
        tc_filter[Tc.Param.SRC_NETWORK] = sanitize_network(
            self.__filter_src_network, self.__ip_version
        )
        tc_filter[Tc.Param.DST_NETWORK] = sanitize_network(
            self.__filter_dst_network, self.__ip_version
        )
        tc_filter[Tc.Param.SRC_PORT] = self.__filter_src_port
        tc_filter[Tc.Param.DST_PORT] = self.__filter_dst_port

        return tc_filter

    def __parse_flow_id(self, line):
        parsed_list = self.__FILTER_FLOWID_PATTERN.parseString(line)
        self.__flow_id = parsed_list[-1]
        logger.debug("succeed to parse flow id: flow-id={}, line={}".format(self.__flow_id, line))

    def __parse_protocol(self, line):
        parsed_list = self.__FILTER_PROTOCOL_PATTERN.parseString(line)
        self.__protocol = parsed_list[-1]
        logger.debug(
            "succeed to parse protocol: protocol={}, line={}".format(self.__protocol, line)
        )

    def __parse_priority(self, line):
        parsed_list = self.__FILTER_PRIORITY_PATTERN.parseString(line)
        self.__priority = int(parsed_list[-1])
        logger.debug(
            "succeed to parse priority: priority={}, line={}".format(self.__priority, line)
        )

    def __parse_filter_id(self, line):
        parsed_list = self.__FILTER_ID_PATTERN.parseString(line)
        self.__filter_id = parsed_list[-1]
        logger.debug(
            "succeed to parse filter id: filter-id={}, line={}".format(self.__filter_id, line)
        )

    def __parse_mangle_mark(self, line):
        parsed_list = self.__FILTER_MANGLE_MARK_PATTERN.parseString(line)
        self.__classid = parsed_list[-1]
        self.__handle = int("0" + parsed_list[-3], 16)
        logger.debug(
            "succeed to parse mangle mark: "
            "classid={}, handle={}, line={}".format(self.__classid, self.__handle, line)
        )

    def __parse_filter_ip_line(self, line):
        parsed_list = self.__FILTER_MATCH_PATTERN.parseString(line)
        value_hex, mask_hex = parsed_list[1].split("/")
        match_id = int(parsed_list[3])

        return (value_hex, mask_hex, match_id)

    def __parse_filter_ipv4_network(self, value_hex, mask_hex, match_id):
        ipaddr = ".".join(
            [text_type(int(value_hex[i : i + 2], 16)) for i in range(0, len(value_hex), 2)]
        )
        netmask = bin(int(mask_hex, 16)).count("1")
        network = "{:s}/{:d}".format(ipaddr, netmask)

        if match_id == self.FilterMatchIdIpv4.INCOMING_NETWORK:
            self.__filter_src_network = network
        elif match_id == self.FilterMatchIdIpv4.OUTGOING_NETWORK:
            self.__filter_dst_network = network
        else:
            logger.warn("unknown match id: {}".format(match_id))

    def __parse_filter_ipv6_network(self, value_hex, mask_hex, match_id):
        from collections import namedtuple

        Ipv6Entry = namedtuple("Ipv6Entry", "match_id octet_list mask_hex")

        OCTET_LEN = 4
        ipv6_entry_list = [
            Ipv6Entry(
                match_id=match_id,
                octet_list=[
                    value_hex[i : i + OCTET_LEN] for i in range(0, len(value_hex), OCTET_LEN)
                ],
                mask_hex=mask_hex,
            )
        ]

        while True:
            try:
                line = self.__buffer[self.__parse_idx].strip()
            except IndexError:
                break

            try:
                value_hex, mask_hex, match_id = self.__parse_filter_ip_line(line)
            except pp.ParseException:
                break

            if (
                match_id in self.FilterMatchIdIpv6.INCOMING_NETWORK_LIST
                or match_id in self.FilterMatchIdIpv6.OUTGOING_NETWORK_LIST
            ):
                ipv6_entry_list.append(
                    Ipv6Entry(
                        match_id=match_id,
                        octet_list=[
                            value_hex[i : i + OCTET_LEN]
                            for i in range(0, len(value_hex), OCTET_LEN)
                        ],
                        mask_hex=mask_hex,
                    )
                )
            else:
                break

            self.__parse_idx += 1

        src_octet_list = []
        dst_octet_list = []
        src_netmask = 0
        dst_netmask = 0

        for ipv6_entry in ipv6_entry_list:
            part_netmask = bin(int(ipv6_entry.mask_hex, 16)).count("1")

            if ipv6_entry.match_id in self.FilterMatchIdIpv6.INCOMING_NETWORK_LIST:
                src_octet_list.extend(ipv6_entry.octet_list)
                src_netmask += part_netmask
            elif ipv6_entry.match_id in self.FilterMatchIdIpv6.OUTGOING_NETWORK_LIST:
                dst_octet_list.extend(ipv6_entry.octet_list)
                dst_netmask += part_netmask
            else:
                raise ValueError("unexpected ipv6 entry: {}".format(ipv6_entry))

        while len(src_octet_list) < 8:
            src_octet_list.append("0000")
        while len(dst_octet_list) < 8:
            dst_octet_list.append("0000")

        self.__filter_dst_network = ipaddress.IPv6Network(
            "{:s}/{:d}".format(":".join(dst_octet_list), dst_netmask)
        ).compressed
        self.__filter_src_network = ipaddress.IPv6Network(
            "{:s}/{:d}".format(":".join(src_octet_list), src_netmask)
        ).compressed

    def __parse_filter_port(self, value_hex):
        # Port filter consists eight hex digits.
        # The upper-half represents source port filter and
        # the bottom-half represents destination port filter.

        if len(value_hex) != 8:
            raise ValueError("invalid port filter value: {}".format(value_hex))

        src_port_hex = value_hex[:4]
        dst_port_hex = value_hex[4:]

        logger.debug(
            "parse ipv4 port: src-port-hex={}, dst-port-hex={}".format(src_port_hex, dst_port_hex)
        )

        src_port_decimal = int(src_port_hex, 16)
        self.__filter_src_port = src_port_decimal if src_port_decimal != 0 else None

        dst_port_decimal = int(dst_port_hex, 16)
        self.__filter_dst_port = dst_port_decimal if dst_port_decimal != 0 else None

    def __parse_filter_ipv4(self, line):
        value_hex, mask_hex, match_id = self.__parse_filter_ip_line(line)

        if match_id in [
            self.FilterMatchIdIpv4.INCOMING_NETWORK,
            self.FilterMatchIdIpv4.OUTGOING_NETWORK,
        ]:
            self.__parse_filter_ipv4_network(value_hex, mask_hex, match_id)
        elif match_id == self.FilterMatchIdIpv4.PORT:
            self.__parse_filter_port(value_hex)
        elif match_id in (
            self.FilterMatchIdIpv6.INCOMING_NETWORK_LIST
            + self.FilterMatchIdIpv6.OUTGOING_NETWORK_LIST
            + [self.FilterMatchIdIpv6.PORT]
        ):
            logger.warn(
                "unknown match id for an IPv4 filter: might be an IPv6 filter. "
                "try to use --ipv6 option. (id={})".format(match_id)
            )
            return
        else:
            logger.debug("unknown match id: {}".format(match_id))
            return

        logger.debug(
            "succeed to parse ipv4 filter: "
            + ", ".join(
                [
                    "src_network={}".format(self.__filter_src_network),
                    "dst_network={}".format(self.__filter_dst_network),
                    "src_port={}".format(self.__filter_src_port),
                    "dst_port={}".format(self.__filter_dst_port),
                    "line={}".format(line),
                ]
            )
        )

    def __parse_filter_ipv6(self, line):
        value_hex, mask_hex, match_id = self.__parse_filter_ip_line(line)

        if (
            match_id in self.FilterMatchIdIpv6.INCOMING_NETWORK_LIST
            or match_id in self.FilterMatchIdIpv6.OUTGOING_NETWORK_LIST
        ):
            self.__parse_filter_ipv6_network(value_hex, mask_hex, match_id)
        elif match_id == self.FilterMatchIdIpv6.PORT:
            self.__parse_filter_port(value_hex)
        else:
            logger.debug("unknown match id: {}".format(match_id))
            return

        logger.debug(
            "succeed to parse ipv6 filter: "
            + ", ".join(
                [
                    "src_network={}".format(self.__filter_src_network),
                    "dst_network={}".format(self.__filter_dst_network),
                    "src_port={}".format(self.__filter_src_port),
                    "dst_port={}".format(self.__filter_dst_port),
                    "line={}".format(line),
                ]
            )
        )



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import humanreadable as hr
import six
import typepy
from pyroute2 import IPRoute

from ._const import Network
from ._error import NetworkInterfaceNotFoundError


def get_anywhere_network(ip_version):
    ip_version_n = typepy.Integer(ip_version).try_convert()

    if ip_version_n == 4:
        return Network.Ipv4.ANYWHERE

    if ip_version_n == 6:
        return Network.Ipv6.ANYWHERE

    raise ValueError("unknown ip version: {}".format(ip_version))


def _get_iproute2_upper_limite_rate():
    """
    :return: Upper bandwidth rate limit of iproute2 [Kbps].
    :rtype: int
    """

    # upper bandwidth rate limit of iproute2 was 34,359,738,360
    # bits per second older than 3.14.0
    # http://git.kernel.org/cgit/linux/kernel/git/shemminger/iproute2.git/commit/?id=8334bb325d5178483a3063c5f06858b46d993dc7

    return hr.BitPerSecond("32Gbps")


def _read_iface_speed(tc_device):
    with open("/sys/class/net/{:s}/speed".format(tc_device)) as f:
        return int(f.read().strip())


def get_upper_limit_rate(tc_device):
    if typepy.is_null_string(tc_device):
        return _get_iproute2_upper_limite_rate()

    try:
        speed_value = _read_iface_speed(tc_device)
    except IOError:
        return _get_iproute2_upper_limite_rate()

    if speed_value < 0:
        # default to the iproute2 upper limit when speed value is -1 in
        # paravirtualized network interfaces
        return _get_iproute2_upper_limite_rate()

    return min(hr.BitPerSecond("{}Mbps".format(speed_value)), _get_iproute2_upper_limite_rate())


def is_anywhere_network(network, ip_version):
    try:
        network = network.strip()
    except AttributeError as e:
        raise ValueError(e)

    if ip_version == 4:
        return network == get_anywhere_network(ip_version)

    if ip_version == 6:
        return network in (get_anywhere_network(ip_version), "0:0:0:0:0:0:0:0/0")

    raise ValueError("invalid ip version: {}".format(ip_version))


def sanitize_network(network, ip_version):
    """
    :return: Network string
    :rtype: str
    :raises ValueError: if the network string is invalid.
    """

    import ipaddress

    if typepy.is_null_string(network) or network.lower() == "anywhere":
        return get_anywhere_network(ip_version)

    try:
        if ip_version == 4:
            ipaddress.IPv4Address(network)
            return network + "/32"

        if ip_version == 6:
            return ipaddress.IPv6Address(network).compressed
    except ipaddress.AddressValueError:
        pass

    # validate network str ---

    if ip_version == 4:
        return ipaddress.IPv4Network(six.text_type(network)).compressed

    if ip_version == 6:
        return ipaddress.IPv6Network(six.text_type(network)).compressed

    raise ValueError("unexpected ip version: {}".format(ip_version))


def verify_network_interface(device, tc_command_output):
    from ._common import is_execute_tc_command

    if not is_execute_tc_command(tc_command_output):
        return

    with IPRoute() as ipr:
        avail_interfaces = [link.get_attr("IFLA_IFNAME") for link in ipr.get_links()]

    if device not in avail_interfaces:
        raise NetworkInterfaceNotFoundError(target=device)



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, print_function

import errno
import sys

import logbook
import subprocrunner as spr

from .__version__ import __version__
from ._argparse_wrapper import ArgparseWrapper
from ._capabilities import check_execution_authority
from ._common import initialize_cli, is_execute_tc_command, normalize_tc_value
from ._const import Tc, TcSubCommand
from ._error import NetworkInterfaceNotFoundError
from ._logger import logger, set_logger
from ._main import Main
from ._network import verify_network_interface
from .traffic_control import TrafficControl


def parse_option():
    parser = ArgparseWrapper(__version__)

    group = parser.parser.add_argument_group("Traffic Control")
    if set(["-d", "--device"]).intersection(set(sys.argv)):
        # deprecated: remain for backward compatibility
        group.add_argument("-d", "--device", required=True, help="network device name (e.g. eth0)")
    else:
        group.add_argument("device", help="network device name (e.g. eth0)")
    group.add_argument(
        "-a",
        "--all",
        dest="is_delete_all",
        action="store_true",
        help="delete all of the shaping rules.",
    )
    group.add_argument(
        "--id",
        dest="filter_id",
        help="""delete a shaping rule which has a specific id. you can get an id (filter_id)
        by tcshow command output.
        e.g. "filter_id": "800::801"
        """,
    )

    parser.add_routing_group()
    parser.add_docker_group()

    return parser.parser.parse_args()


class TcDelMain(Main):
    def run(self, is_delete_all):
        return_code_list = []

        for tc_target in self._fetch_tc_targets():
            tc = self.__create_tc_obj(tc_target)
            if self._options.log_level == logbook.INFO:
                spr.set_log_level(logbook.ERROR)
            normalize_tc_value(tc)

            try:
                if is_delete_all:
                    return_code_list.append(tc.delete_all_rules())
                else:
                    return_code_list.append(tc.delete_tc())
            except NetworkInterfaceNotFoundError as e:
                logger.error(e)
                return errno.EINVAL

            self._dump_history(tc, Tc.Command.TCDEL)

        return self._get_return_code(return_code_list)

    def __create_tc_obj(self, tc_target):
        from .parser.shaping_rule import TcShapingRuleParser
        from simplesqlite.query import Where

        options = self._options

        if options.filter_id:
            ip_version = 6 if options.is_ipv6 else 4
            shaping_rule_parser = TcShapingRuleParser(
                device=tc_target,
                ip_version=ip_version,
                tc_command_output=options.tc_command_output,
                logger=logger,
            )
            shaping_rule_parser.parse()
            result = shaping_rule_parser.con.select_as_dict(
                table_name=TcSubCommand.FILTER.value,
                columns=[
                    Tc.Param.SRC_NETWORK,
                    Tc.Param.DST_NETWORK,
                    Tc.Param.SRC_PORT,
                    Tc.Param.DST_PORT,
                ],
                where=Where(Tc.Param.FILTER_ID, options.filter_id),
            )
            if not result:
                logger.error(
                    "shaping rule not found associated with the id ({}).".format(options.filter_id)
                )
                sys.exit(1)

            filter_param = result[0]
            dst_network = filter_param.get(Tc.Param.DST_NETWORK)
            src_network = filter_param.get(Tc.Param.SRC_NETWORK)
            dst_port = filter_param.get(Tc.Param.DST_PORT)
            src_port = filter_param.get(Tc.Param.SRC_PORT)
        else:
            dst_network = self._extract_dst_network()
            src_network = self._extract_src_network()
            dst_port = options.dst_port
            src_port = options.src_port

        return TrafficControl(
            tc_target,
            direction=options.direction,
            dst_network=dst_network,
            src_network=src_network,
            dst_port=dst_port,
            src_port=src_port,
            is_ipv6=options.is_ipv6,
            tc_command_output=options.tc_command_output,
        )


def main():
    options = parse_option()

    initialize_cli(options)

    if is_execute_tc_command(options.tc_command_output):
        check_execution_authority("tc")

        if not options.use_docker:
            try:
                verify_network_interface(options.device, options.tc_command_output)
            except NetworkInterfaceNotFoundError as e:
                logger.error(e)
                return errno.EINVAL

        is_delete_all = options.is_delete_all
    else:
        spr.SubprocessRunner.default_is_dry_run = True
        is_delete_all = True
        set_logger(False)

    spr.SubprocessRunner.clear_history()

    return TcDelMain(options).run(is_delete_all)


if __name__ == "__main__":
    sys.exit(main())



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import print_function, unicode_literals

import pytest
import six
from simplesqlite import connect_memdb

import tcconfig.parser._filter
import tcconfig.parser._qdisc
import tcconfig.parser.shaping_rule
from tcconfig._const import Tc
from tcconfig.parser._class import TcClassParser

from .common import print_test_result


DEVICE = "eth0"


@pytest.fixture
def filter_parser_ipv4():
    return tcconfig.parser._filter.TcFilterParser(connect_memdb(), ip_version=4)


@pytest.fixture
def filter_parser_ipv6():
    return tcconfig.parser._filter.TcFilterParser(connect_memdb(), ip_version=6)


@pytest.fixture
def qdisc_parser():
    return tcconfig.parser._qdisc.TcQdiscParser(connect_memdb())


@pytest.fixture
def class_parser():
    return TcClassParser(connect_memdb())


class Test_TcFilterParser_parse_filter_ipv4(object):
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            [None, []],
            ["", []],
            [
                six.b(
                    """filter parent 1: protocol ip pref 1 u32
filter parent 1: protocol ip pref 1 u32 fh 801: ht divisor 1
filter parent 1: protocol ip pref 1 u32 fh 801::800 order 2048 key ht 801 bkt 0 flowid 1:1
  match c0a8000a/ffffffff at 16
filter parent 1: protocol ip pref 2 u32
filter parent 1: protocol ip pref 2 u32 fh 800: ht divisor 1
filter parent 1: protocol ip pref 2 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1:2
  match 00000000/00000000 at 16"""
                ),
                [
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.FILTER_ID: "801::800",
                        Tc.Param.FLOW_ID: "1:1",
                        Tc.Param.SRC_NETWORK: "0.0.0.0/0",
                        Tc.Param.DST_NETWORK: "192.168.0.10/32",
                        Tc.Param.PROTOCOL: "ip",
                        Tc.Param.PRIORITY: 1,
                        Tc.Param.SRC_PORT: None,
                        Tc.Param.DST_PORT: None,
                    },
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.FILTER_ID: "800::800",
                        Tc.Param.FLOW_ID: "1:2",
                        Tc.Param.SRC_NETWORK: "0.0.0.0/0",
                        Tc.Param.DST_NETWORK: "0.0.0.0/0",
                        Tc.Param.PROTOCOL: "ip",
                        Tc.Param.PRIORITY: 2,
                        Tc.Param.SRC_PORT: None,
                        Tc.Param.DST_PORT: None,
                    },
                ],
            ],
            [
                six.b(
                    """filter parent 1: protocol ip pref 1 u32
filter parent 1: protocol ip pref 1 u32 fh 801: ht divisor 1
filter parent 1: protocol ip pref 1 u32 fh 801::800 order 2048 key ht 801 bkt 0 flowid 1:1
  match c0a80000/ffffff00 at 16
  match 00000050/0000ffff at 20
filter parent 1: protocol ip pref 2 u32
filter parent 1: protocol ip pref 2 u32 fh 800: ht divisor 1
filter parent 1: protocol ip pref 2 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1:2
  match 00000000/00000000 at 16
  match 04d20000/ffff0000 at 20"""
                ),
                [
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.FILTER_ID: "801::800",
                        Tc.Param.FLOW_ID: "1:1",
                        Tc.Param.SRC_NETWORK: "0.0.0.0/0",
                        Tc.Param.DST_NETWORK: "192.168.0.0/24",
                        Tc.Param.PROTOCOL: "ip",
                        Tc.Param.PRIORITY: 1,
                        Tc.Param.SRC_PORT: None,
                        Tc.Param.DST_PORT: 80,
                    },
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.FILTER_ID: "800::800",
                        Tc.Param.FLOW_ID: "1:2",
                        Tc.Param.SRC_NETWORK: "0.0.0.0/0",
                        Tc.Param.DST_NETWORK: "0.0.0.0/0",
                        Tc.Param.PROTOCOL: "ip",
                        Tc.Param.PRIORITY: 2,
                        Tc.Param.SRC_PORT: 1234,
                        Tc.Param.DST_PORT: None,
                    },
                ],
            ],
            [
                six.b(
                    """filter parent 1: protocol ip pref 1 u32
filter parent 1: protocol ip pref 1 u32 fh 801: ht divisor 1
filter parent 1: protocol ip pref 1 u32 fh 801::800 order 2048 key ht 801 bkt 0 flowid 1:3
  match c0a8000a/ffffffff at 12
  match 00001f90/0000ffff at 20
filter parent 1: protocol ip pref 2 u32
filter parent 1: protocol ip pref 2 u32 fh 800: ht divisor 1
filter parent 1: protocol ip pref 2 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1:2
  match 00000000/00000000 at 12"""
                ),
                [
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.FILTER_ID: "801::800",
                        Tc.Param.FLOW_ID: "1:3",
                        Tc.Param.SRC_NETWORK: "192.168.0.10/32",
                        Tc.Param.DST_NETWORK: "0.0.0.0/0",
                        Tc.Param.PROTOCOL: "ip",
                        Tc.Param.PRIORITY: 1,
                        Tc.Param.SRC_PORT: None,
                        Tc.Param.DST_PORT: 8080,
                    },
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.FILTER_ID: "800::800",
                        Tc.Param.FLOW_ID: "1:2",
                        Tc.Param.SRC_NETWORK: "0.0.0.0/0",
                        Tc.Param.DST_NETWORK: "0.0.0.0/0",
                        Tc.Param.PROTOCOL: "ip",
                        Tc.Param.PRIORITY: 2,
                        Tc.Param.SRC_PORT: None,
                        Tc.Param.DST_PORT: None,
                    },
                ],
            ],
            [
                six.b(
                    """filter parent 1a1a: protocol ip pref 1 u32
filter parent 1a1a: protocol ip pref 1 u32 fh 800: ht divisor 1
filter parent 1a1a: protocol ip pref 1 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1a1a:2
  match 00000000/00000000 at 16
  match 15b3115c/ffffffff at 20"""
                ),
                [
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.FILTER_ID: "800::800",
                        Tc.Param.FLOW_ID: "1a1a:2",
                        Tc.Param.SRC_NETWORK: "0.0.0.0/0",
                        Tc.Param.DST_NETWORK: "0.0.0.0/0",
                        Tc.Param.PROTOCOL: "ip",
                        Tc.Param.PRIORITY: 1,
                        Tc.Param.SRC_PORT: 5555,
                        Tc.Param.DST_PORT: 4444,
                    }
                ],
            ],
            [
                six.b(
                    """filter parent 1f1c: protocol ip pref 1 fw
filter parent 1f1c: protocol ip pref 1 fw handle 0x65 classid 1f1c:1"""
                ),
                [{Tc.Param.DEVICE: DEVICE, "classid": "1f1c:1", "handle": 101}],
            ],
        ],
    )
    def test_normal(self, filter_parser_ipv4, value, expected):
        actual = filter_parser_ipv4.parse(DEVICE, value)

        print_test_result(expected=expected, actual=actual)

        assert actual == expected


class Test_TcFilterParser_parse_filter_ipv6(object):
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            [None, []],
            ["", []],
            [
                six.b(
                    """filter parent 1f87: protocol ipv6 pref 1 u32
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800: ht divisor 1
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1f87:2
  match 00000000/ffffffff at 24
  match 00000000/ffffffff at 28
  match 00000000/ffffffff at 32
  match 00000001/ffffffff at 36"""
                ),
                [
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.FILTER_ID: "800::800",
                        Tc.Param.FLOW_ID: "1f87:2",
                        Tc.Param.PROTOCOL: "ipv6",
                        Tc.Param.PRIORITY: 1,
                        Tc.Param.SRC_NETWORK: "::/0",
                        Tc.Param.DST_NETWORK: "::1/128",
                        Tc.Param.SRC_PORT: None,
                        Tc.Param.DST_PORT: None,
                    }
                ],
            ],
            [
                six.b(
                    """filter parent 1f87: protocol ipv6 pref 1 u32
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800: ht divisor 1
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1f87:2
  match 2001db00/ffffffff at 8
  match 00000000/ffffffff at 12
  match 00000000/ffffffff at 16
  match 00000001/ffffffff at 20"""
                ),
                [
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.FILTER_ID: "800::800",
                        Tc.Param.FLOW_ID: "1f87:2",
                        Tc.Param.PROTOCOL: "ipv6",
                        Tc.Param.PRIORITY: 1,
                        Tc.Param.SRC_NETWORK: "2001:db00::1/128",
                        Tc.Param.DST_NETWORK: "::/0",
                        Tc.Param.SRC_PORT: None,
                        Tc.Param.DST_PORT: None,
                    }
                ],
            ],
            [
                six.b(
                    """filter parent 1f87: protocol ipv6 pref 1 u32
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800: ht divisor 1
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1f87:2
  match 2001db00/ffffff00 at 24
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800::801 order 2049 key ht 800 bkt 0 flowid 1f87:3
  match 2001db00/ffffffff at 24
  match 00000000/ffffffff at 28
  match 00000000/ffffffff at 32
  match 00000001/ffffffff at 36
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800::802 order 2050 key ht 800 bkt 0 flowid 1f87:4
  match 00001f90/0000ffff at 40"""
                ),
                [
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.FILTER_ID: "800::800",
                        Tc.Param.FLOW_ID: "1f87:2",
                        Tc.Param.PROTOCOL: "ipv6",
                        Tc.Param.PRIORITY: 1,
                        Tc.Param.SRC_NETWORK: "::/0",
                        Tc.Param.DST_NETWORK: "2001:db00::/24",
                        Tc.Param.SRC_PORT: None,
                        Tc.Param.DST_PORT: None,
                    },
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.FILTER_ID: "800::801",
                        Tc.Param.FLOW_ID: "1f87:3",
                        Tc.Param.PROTOCOL: "ipv6",
                        Tc.Param.PRIORITY: 1,
                        Tc.Param.SRC_NETWORK: "::/0",
                        Tc.Param.DST_NETWORK: "2001:db00::1/128",
                        Tc.Param.SRC_PORT: None,
                        Tc.Param.DST_PORT: None,
                    },
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.FILTER_ID: "800::802",
                        Tc.Param.FLOW_ID: "1f87:4",
                        Tc.Param.PROTOCOL: "ipv6",
                        Tc.Param.PRIORITY: 1,
                        Tc.Param.SRC_NETWORK: "::/0",
                        Tc.Param.DST_NETWORK: "::/0",
                        Tc.Param.SRC_PORT: None,
                        Tc.Param.DST_PORT: 8080,
                    },
                ],
            ],
            [
                six.b(
                    """filter parent 1f87: protocol ipv6 pref 1 u32
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800: ht divisor 1
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800::802 order 2050 key ht 800 bkt 0 flowid 1f87:4
  match 00501f90/ffffffff at 40"""
                ),
                [
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.FILTER_ID: "800::802",
                        Tc.Param.FLOW_ID: "1f87:4",
                        Tc.Param.PROTOCOL: "ipv6",
                        Tc.Param.PRIORITY: 1,
                        Tc.Param.SRC_NETWORK: "::/0",
                        Tc.Param.DST_NETWORK: "::/0",
                        Tc.Param.SRC_PORT: 80,
                        Tc.Param.DST_PORT: 8080,
                    }
                ],
            ],
        ],
    )
    def test_normal(self, filter_parser_ipv6, value, expected):
        actual = filter_parser_ipv6.parse(DEVICE, value)

        print_test_result(expected=expected, actual=actual)

        assert actual == expected


class Test_TcFilterParser_parse_incoming_device(object):
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            ["", None],
            [None, None],
            [
                six.b(
                    """filter parent ffff: protocol ip pref 49152 u32
filter parent ffff: protocol ip pref 49152 u32 fh 800: ht divisor 1
filter parent ffff: protocol ip pref 49152 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1f87:
  match 00000000/00000000 at 0
        action order 1: mirred (Egress Redirect to device ifb8071) stolen
        index 98 ref 1 bind 1"""
                ),
                "ifb8071",
            ],
            [
                six.b(
                    """filter parent ffff: protocol ip pref 49152 u32
filter parent ffff: protocol ip pref 49152 u32 fh 800: ht divisor 1
filter parent ffff: protocol ip pref 49152 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1:
  match 00000000/00000000 at 0
        action order 1: mirred (Egress Redirect to device ifb0) stolen
        index 3795 ref 1 bind 1"""
                ),
                "ifb0",
            ],
        ],
    )
    def test_normal(self, filter_parser_ipv4, value, expected):
        assert filter_parser_ipv4.parse_incoming_device(value) == expected


class Test_TcQdiscParser_parse(object):
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            [
                six.b(
                    """qdisc htb 1f87: root refcnt 2 r2q 10 default 1 direct_packets_stat 1 direct_qlen 1000
qdisc netem 2007: parent 1f87:2 limit 1000 delay 1.0ms loss 0.01%
"""
                ),
                [
                    {
                        Tc.Param.DEVICE: DEVICE,
                        "delay": "1.0ms",
                        "loss": "0.01%",
                        Tc.Param.HANDLE: "2007:",
                        Tc.Param.PARENT: "1f87:2",
                        "direct_qlen": 1000,
                    }
                ],
            ],
            [
                six.b(
                    """qdisc htb 1a1a: root refcnt 2 r2q 10 default 1 direct_packets_stat 4 direct_qlen 1000
qdisc netem 1a9a: parent 1a1a:2 limit 1000
"""
                ),
                [
                    {
                        Tc.Param.DEVICE: DEVICE,
                        Tc.Param.HANDLE: "1a9a:",
                        Tc.Param.PARENT: "1a1a:2",
                        "direct_qlen": 1000,
                    }
                ],
            ],
            [
                six.b(
                    """
qdisc htb 1f87: root refcnt 2 r2q 10 default 1 direct_packets_stat 0 direct_qlen 1000
qdisc netem 2007: parent 1f87:2 limit 1000 delay 5.0ms
qdisc netem 2008: parent 1f87:3 limit 1000 delay 50.0ms  1.0ms loss 5%
"""
                ),
                [
                    {
                        Tc.Param.DEVICE: DEVICE,
                        "delay": "5.0ms",
                        Tc.Param.HANDLE: "2007:",
                        Tc.Param.PARENT: "1f87:2",
                        "direct_qlen": 1000,
                    },
                    {
                        Tc.Param.DEVICE: DEVICE,
                        "delay": "50.0ms",
                        "loss": "5%",
                        "delay-distro": "1.0ms",
                        Tc.Param.HANDLE: "2008:",
                        Tc.Param.PARENT: "1f87:3",
                    },
                ],
            ],
            [
                six.b(
                    """
qdisc htb 1f87: root refcnt 2 r2q 10 default 1 direct_packets_stat 0 direct_qlen 1000
qdisc netem 2007: parent 1f87:2 limit 1000 delay 2.5s
qdisc netem 2008: parent 1f87:3 limit 1000 delay 0.5s  1.0ms loss 5%
"""
                ),
                [
                    {
                        Tc.Param.DEVICE: DEVICE,
                        "delay": "2.5s",
                        Tc.Param.HANDLE: "2007:",
                        Tc.Param.PARENT: "1f87:2",
                        "direct_qlen": 1000,
                    },
                    {
                        Tc.Param.DEVICE: DEVICE,
                        "delay": "0.5s",
                        "loss": "5%",
                        "delay-distro": "1.0ms",
                        Tc.Param.HANDLE: "2008:",
                        Tc.Param.PARENT: "1f87:3",
                    },
                ],
            ],
        ],
    )
    def test_normal(self, qdisc_parser, value, expected):
        actual = qdisc_parser.parse(DEVICE, value)

        import json

        print_test_result(
            expected=json.dumps(expected, indent=4), actual=json.dumps(actual, indent=4)
        )

        assert actual == expected


class Test_TcClassParser_parse(object):
    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            [
                six.b(
                    """
class htb 120a:1 root prio rate 32Gbit ceil 32Gbit burst 0b cburst 0b
class htb 120a:2 root leaf 2946: prio rate 1Gbit ceil 1Gbit burst 125000Kb cburst 125000Kb
"""
                ),
                [
                    {
                        TcClassParser.Key.RATE: "32Gbps",
                        TcClassParser.Key.CLASS_ID: "120a:1",
                        TcClassParser.Key.DEVICE: "eth0",
                    },
                    {
                        TcClassParser.Key.RATE: "1Gbps",
                        TcClassParser.Key.CLASS_ID: "120a:2",
                        TcClassParser.Key.DEVICE: "eth0",
                    },
                ],
            ],
            [
                six.b(
                    """
class htb 120a:1 root prio rate 32Gbit ceil 32Gbit burst 0b cburst 0b
class htb 120a:2 root leaf 2518: prio rate 200Kbit ceil 200Kbit burst 25Kb cburst 25Kb
                    """
                ),
                [
                    {
                        TcClassParser.Key.RATE: "32Gbps",
                        TcClassParser.Key.DEVICE: "eth0",
                        TcClassParser.Key.CLASS_ID: "120a:1",
                    },
                    {
                        TcClassParser.Key.RATE: "200Kbps",
                        TcClassParser.Key.DEVICE: "eth0",
                        TcClassParser.Key.CLASS_ID: "120a:2",
                    },
                ],
            ],
        ],
    )
    def test_normal(self, class_parser, value, expected):
        actual = class_parser.parse(DEVICE, value)

        import json

        print_test_result(
            expected=json.dumps(expected, indent=4), actual=json.dumps(actual, indent=4)
        )

        assert actual == expected



# encoding: utf-8

from __future__ import absolute_import, division, print_function

import pytest
import simplejson as json

import tcconfig
from tcconfig._const import TrafficDirection
from tcconfig._netem_param import NetemParameter
from tcconfig.traffic_control import TrafficControl, delete_all_rules

from ..common import NullLogger, print_test_result


@pytest.fixture
def device_value(request):
    return request.config.getoption("--device")


class Test_TcShapingRuleParser(object):
    def test_normal(self, device_value):
        if device_value is None:
            pytest.skip("device option is null")

        delete_all_rules(device_value)

        tc = TrafficControl(
            device=device_value,
            shaping_algorithm="htb",
            direction=TrafficDirection.INCOMING,
            netem_param=NetemParameter(
                device=device_value,
                bandwidth_rate="200kbps",
                packet_loss_rate=0,
                packet_duplicate_rate=0,
                corruption_rate=0,
                reordering_rate=0,
            ),
            src_network="192.168.3.188",
            dst_port=5201,
        )
        assert tc.set_shaping_rule() == 0

        tc = TrafficControl(
            device=device_value,
            shaping_algorithm="htb",
            direction=TrafficDirection.OUTGOING,
            netem_param=NetemParameter(
                device=device_value,
                bandwidth_rate="1000000kbps",
                packet_loss_rate=0,
                packet_duplicate_rate=0,
                corruption_rate=0,
                reordering_rate=0,
            ),
            dst_network="192.168.3.188",
        )
        assert tc.set_shaping_rule() == 0

        result = tcconfig.parser.shaping_rule.TcShapingRuleParser(
            device=device_value,
            ip_version=4,
            logger=NullLogger(),
            tc_command_output=None,
            is_parse_filter_id=False,
        ).get_tc_parameter()
        print_test_result(expected={}, actual=result)
        assert result == json.loads(
            """\
            {
                "enp0s3": {
                    "outgoing": {
                        "dst-network=192.168.3.188/32, protocol=ip": {
                            "rate": "1Gbps"
                        }
                    },
                    "incoming": {
                        "src-network=192.168.3.188/32, dst-port=5201, protocol=ip": {
                            "rate": "200Kbps"
                        }
                    }
                }
            }
            """
        )



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import errno
import re

import typepy
from subprocrunner import SubprocessRunner
from typepy import Integer

from ._common import find_bin_path
from ._const import LIST_MANGLE_TABLE_OPTION, Network
from ._logger import logger
from ._network import sanitize_network
from ._split_line_list import split_line_list


VALID_CHAIN_LIST = ["PREROUTING", "INPUT", "OUTPUT"]


def get_iptables_base_command():
    iptables_path = find_bin_path("iptables")

    if iptables_path:
        if re.search("iptables$", iptables_path):
            return iptables_path

        # debian/ubuntu may return /sbin/xtables-multi
        return "{:s} iptables".format(iptables_path)

    return None


class IptablesMangleMarkEntry(object):
    @property
    def line_number(self):
        return self.__line_number

    @property
    def protocol(self):
        return self.__protocol

    @property
    def source(self):
        return self.__source

    @property
    def destination(self):
        return self.__destination

    @property
    def mark_id(self):
        return self.__mark_id

    @property
    def chain(self):
        return self.__chain

    def __init__(
        self, ip_version, mark_id, source, destination, chain, protocol="all", line_number=None
    ):
        self.__line_number = line_number
        self.__mark_id = mark_id
        self.__source = sanitize_network(source, ip_version)
        self.__destination = sanitize_network(destination, ip_version)
        self.__protocol = protocol

        if chain not in VALID_CHAIN_LIST:
            raise ValueError("invalid chain: {}".format(chain))

        self.__chain = chain

    def __eq__(self, other):
        return all(
            [
                self.chain == other.chain,
                self.mark_id == other.mark_id,
                self.protocol == other.protocol,
                self.source == other.source,
                self.destination == other.destination,
            ]
        )

    def __repr__(self, *args, **kwargs):
        str_list = []

        if Integer(self.line_number).is_type():
            str_list.append("line-num={}".format(self.line_number))

        str_list.extend(
            [
                "protocol={:s}".format(self.protocol),
                "source={:s}".format(self.source),
                "destination={:s}".format(self.destination),
                "mark_id={:d}".format(self.mark_id),
                "chain={:s}".format(self.chain),
            ]
        )

        return ", ".join(str_list)

    def to_append_command(self):
        Integer(self.mark_id).validate()

        command_item_list = [
            "{:s} -A {:s} -t mangle -j MARK".format(get_iptables_base_command(), self.chain),
            "--set-mark {}".format(self.mark_id),
        ]

        if typepy.is_not_null_string(self.protocol) or Integer(self.protocol).is_type():
            command_item_list.append("-p {}".format(self.protocol))
        if self.__is_valid_srcdst(self.source):
            command_item_list.append("-s {:s}".format(self.source))
        if self.__is_valid_srcdst(self.destination):
            command_item_list.append("-d {:s}".format(self.destination))

        return " ".join(command_item_list)

    def to_delete_command(self):
        Integer(self.line_number).validate()

        return "{:s} -t mangle -D {:s} {}".format(
            get_iptables_base_command(), self.chain, self.line_number
        )

    @staticmethod
    def __is_valid_srcdst(srcdst):
        return typepy.is_not_null_string(srcdst) and srcdst not in (
            Network.Ipv4.ANYWHERE,
            Network.Ipv6.ANYWHERE,
        )


class IptablesMangleController(object):

    __RE_CHAIN = re.compile("Chain {:s} |Chain {:s} |Chain {:s} ".format(*VALID_CHAIN_LIST))
    __RE_CHAIN_NAME = re.compile("{:s}|{:s}|{:s}".format(*VALID_CHAIN_LIST))
    __MAX_MARK_ID = 0xFFFFFFFF
    __MARK_ID_OFFSET = 100

    @property
    def enable(self):
        return self.__enable

    def __init__(self, enable, ip_version):
        self.__enable = enable
        self.__ip_version = ip_version

    def clear(self):
        if not self.enable:
            return

        self.__check_execution_authority()

        for mangle in self.parse():
            proc = SubprocessRunner(mangle.to_delete_command())
            if proc.run() != 0:
                raise OSError(proc.returncode, proc.stderr)

    def get_iptables(self):
        self.__check_execution_authority()

        proc = SubprocessRunner(
            "{:s} {:s}".format(get_iptables_base_command(), LIST_MANGLE_TABLE_OPTION)
        )
        if proc.run() != 0:
            raise OSError(proc.returncode, proc.stderr)

        return proc.stdout

    def get_unique_mark_id(self):
        self.__check_execution_authority()

        mark_id_list = [mangle.mark_id for mangle in self.parse()]
        logger.debug("mangle mark list: {}".format(mark_id_list))

        unique_mark_id = 1 + self.__MARK_ID_OFFSET
        while unique_mark_id < self.__MAX_MARK_ID:
            if unique_mark_id not in mark_id_list:
                return unique_mark_id

            unique_mark_id += 1

        raise RuntimeError("usable mark id not found")

    def parse(self):
        self.__check_execution_authority()

        MANGLE_ITEM_COUNT = 6

        for block in split_line_list(self.get_iptables().splitlines()):
            if len(block) <= 1:
                # skip if there is no mangle table entry exists
                continue

            match = self.__RE_CHAIN.search(block[0])
            if match is None:
                continue

            chain = self.__RE_CHAIN_NAME.search(match.group()).group()

            for line in reversed(block[2:]):
                item_list = line.split()
                if len(item_list) < MANGLE_ITEM_COUNT:
                    continue

                line_number = int(item_list[0])
                target = item_list[1]
                protocol = item_list[2]
                source = item_list[4]
                destination = item_list[5]

                try:
                    mark = int(item_list[-1], 16)
                except ValueError:
                    continue

                if target != "MARK":
                    continue

                yield IptablesMangleMarkEntry(
                    ip_version=self.__ip_version,
                    mark_id=mark,
                    source=source,
                    destination=destination,
                    chain=chain,
                    protocol=protocol,
                    line_number=line_number,
                )

    @classmethod
    def add(cls, mangling_mark):
        if not cls.enable:
            return 0

        cls.__check_execution_authority()

        return SubprocessRunner(mangling_mark.to_append_command()).run()

    @staticmethod
    def __check_execution_authority():
        from ._capabilities import get_permission_error_message, has_execution_authority

        if not has_execution_authority("iptables"):
            raise OSError(errno.EPERM, get_permission_error_message("iptables"))



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import datetime
import io
import os
import sys

import typepy

from ._const import Tc
from ._logger import logger


def write_tc_script(tcconfig_command, command_history, filename_suffix=None):
    filename_item_list = [tcconfig_command]
    if typepy.is_not_null_string(filename_suffix):
        filename_item_list.append(filename_suffix)

    script_line_list = ["#!/bin/sh", ""]

    org_tcconfig_cmd = _get_original_tcconfig_command(tcconfig_command)

    if tcconfig_command != Tc.Command.TCSHOW:
        script_line_list.extend(
            [
                "# command sequence in this script attempt to simulate the following "
                "tcconfig command:",
                "#",
                "#   {:s}".format(org_tcconfig_cmd),
            ]
        )

    script_line_list.extend(
        [
            "#",
            "# the script execution result may different from '{}'".format(org_tcconfig_cmd),
            "#",
            "# created by {:s} on {:s}.".format(
                tcconfig_command, datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S%z")
            ),
            "",
            command_history,
        ]
    )

    filename = "_".join(filename_item_list) + ".sh"
    with io.open(filename, "w", encoding="utf8") as fp:
        fp.write("\n".join(script_line_list) + "\n")

    os.chmod(filename, 0o755)
    logger.info("written a tc script to '{:s}'".format(filename))


def _get_original_tcconfig_command(tcconfig_command):
    return " ".join(
        [tcconfig_command]
        + [command_item for command_item in sys.argv[1:] if command_item != "--tc-script"]
    )



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import pytest
import simplejson as json
from subprocrunner import SubprocessRunner

from tcconfig._const import Tc
from tcconfig.traffic_control import delete_all_rules

from .common import print_test_result, runner_helper


@pytest.fixture
def device_value(request):
    return request.config.getoption("--device")


class Test_tcshow(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:

        python setup.py test --addopts "--runxfail --device=<test device>"
    """

    @pytest.mark.parametrize(["colorize_option"], [[""], ["--color"]])
    def test_normal_empty(self, device_value, colorize_option):
        if device_value is None:
            pytest.skip("device option is null")

        for tc_target in [device_value, "--device {}".format(device_value)]:
            delete_all_rules(tc_target)

            runner = SubprocessRunner(" ".join([Tc.Command.TCSHOW, tc_target, colorize_option]))
            expected = (
                "{"
                + '"{:s}"'.format(device_value)
                + ": {"
                + """
                        "outgoing": {
                        },
                        "incoming": {
                        }
                    }
                }"""
            )

            print(runner.command_str)
            runner.run()

            print_test_result(expected=expected, actual=runner.stdout, error=runner.stderr)
            assert runner.returncode == 0
            assert json.loads(runner.stdout) == json.loads(expected)

    @pytest.mark.parametrize(["colorize_option"], [[""], ["--color"]])
    def test_normal_ipv4(self, device_value, colorize_option):
        if device_value is None:
            pytest.skip("device option is null")

        for tc_target in [device_value, "--device {}".format(device_value)]:
            runner_helper(
                " ".join(
                    [
                        Tc.Command.TCSET,
                        tc_target,
                        "--delay",
                        "10",
                        "--delay-distro",
                        "2",
                        "--loss",
                        "0.01",
                        "--duplicate",
                        "0.5",
                        "--reorder",
                        "0.2",
                        "--rate",
                        "0.25Kbps",
                        "--network",
                        "192.168.0.10",
                        "--port",
                        "8080",
                        "--overwrite",
                    ]
                )
            )

            runner_helper(
                " ".join(
                    [
                        Tc.Command.TCSET,
                        tc_target,
                        "--delay",
                        "1",
                        "--loss",
                        "1",
                        "--rate",
                        "100Mbit/s",
                        "--network",
                        "192.168.1.0/24",
                        "--add",
                    ]
                )
            )
            runner_helper(
                " ".join(
                    [
                        Tc.Command.TCSET,
                        tc_target,
                        "--delay",
                        "10",
                        "--delay-distro",
                        "2",
                        "--rate",
                        "500Kbit/s",
                        "--direction",
                        "incoming",
                    ]
                )
            )
            runner_helper(
                " ".join(
                    [
                        Tc.Command.TCSET,
                        tc_target,
                        "--delay",
                        "1",
                        "--loss",
                        "0.02",
                        "--duplicate",
                        "0.5",
                        "--reorder",
                        "0.2",
                        "--rate",
                        "0.1Mbit/s",
                        "--network",
                        "192.168.11.0/24",
                        "--port",
                        "80",
                        "--direction",
                        "incoming",
                        "--add",
                    ]
                )
            )

            runner = SubprocessRunner(" ".join([Tc.Command.TCSHOW, tc_target, colorize_option]))
            expected = (
                "{"
                + '"{:s}"'.format(device_value)
                + ": {"
                + """
                        "outgoing": {
                        "dst-network=192.168.0.10/32, dst-port=8080, protocol=ip": {
                                "filter_id": "800::800",
                                "delay": "10.0ms",
                                "loss": "0.01%",
                                "duplicate": "0.5%",
                                "reorder": "0.2%",
                                "rate": "248bps",
                                "delay-distro": "2.0ms"
                            },
                            "dst-network=192.168.1.0/24, protocol=ip": {
                                "filter_id": "800::801",
                                "delay": "1.0ms",
                                "loss": "1%",
                                "rate": "100Mbps"
                            }
                        },
                        "incoming": {
                            "dst-network=192.168.11.0/24, dst-port=80, protocol=ip": {
                                "filter_id": "800::801",
                                "delay": "1.0ms",
                                "loss": "0.02%",
                                "duplicate": "0.5%",
                                "reorder": "0.2%",
                                "rate": "100Kbps"
                            },
                            "protocol=ip": {
                                "filter_id": "800::800",
                                "delay": "10.0ms",
                                "delay-distro": "2.0ms",
                                "rate": "500Kbps"
                            }
                        }
                    }
                }"""
            )

            runner.run()
            print_test_result(expected=expected, actual=runner.stdout, error=runner.stderr)

            assert json.loads(runner.stdout) == json.loads(expected)

            delete_all_rules(tc_target)

    @pytest.mark.parametrize(["colorize_option"], [[""], ["--color"]])
    def test_normal_ipv6(self, device_value, colorize_option):
        if device_value is None:
            pytest.skip("device option is null")

        for tc_target in [device_value, "--device {}".format(device_value)]:
            runner_helper(
                " ".join(
                    [
                        Tc.Command.TCSET,
                        tc_target,
                        "--delay",
                        "10",
                        "--delay-distro",
                        "2",
                        "--loss",
                        "0.01",
                        "--duplicate",
                        "5",
                        "--reorder",
                        "2",
                        "--rate",
                        "0.25Kbps",
                        "--network",
                        "::1",
                        "--port",
                        "8080",
                        "--overwrite",
                        "--ipv6",
                    ]
                )
            )
            runner_helper(
                " ".join(
                    [
                        Tc.Command.TCSET,
                        tc_target,
                        "--delay",
                        "1",
                        "--loss",
                        "1",
                        "--rate",
                        "100Mbit/s",
                        "--network",
                        "2001:db00::0/24",
                        "--add",
                        "--ipv6",
                    ]
                )
            )
            runner_helper(
                " ".join(
                    [
                        Tc.Command.TCSET,
                        tc_target,
                        "--delay",
                        "10",
                        "--delay-distro",
                        "2",
                        "--rate",
                        "500Kbit/s",
                        "--direction",
                        "incoming",
                        "--ipv6",
                    ]
                )
            )
            runner_helper(
                " ".join(
                    [
                        Tc.Command.TCSET,
                        tc_target,
                        "--delay",
                        "1",
                        "--loss",
                        "0.02",
                        "--duplicate",
                        "5",
                        "--reorder",
                        "2",
                        "--rate",
                        "0.1Mbit/s",
                        "--network",
                        "2001:db00::0/25",
                        "--port",
                        "80",
                        "--direction",
                        "incoming",
                        "--add",
                        "--ipv6",
                    ]
                )
            )

            runner = SubprocessRunner(
                " ".join([Tc.Command.TCSHOW, tc_target, "--ipv6", colorize_option])
            )
            expected = (
                "{"
                + '"{:s}"'.format(device_value)
                + ": {"
                + """
                        "outgoing": {
                            "dst-network=::1/128, dst-port=8080, protocol=ipv6": {
                                "filter_id": "800::800",
                                "delay": "10.0ms",
                                "loss": "0.01%",
                                "duplicate": "5%",
                                "reorder": "2%",
                                "rate": "248bps",
                                "delay-distro": "2.0ms"
                            },
                            "dst-network=2001:db00::/24, protocol=ipv6": {
                                "filter_id": "800::801",
                                "delay": "1.0ms",
                                "loss": "1%",
                                "rate": "100Mbps"
                            }
                        },
                        "incoming": {
                            "dst-network=2001:db00::/25, dst-port=80, protocol=ipv6": {
                                "filter_id": "800::801",
                                "delay": "1.0ms",
                                "loss": "0.02%",
                                "duplicate": "5%",
                                "reorder": "2%",
                                "rate": "100Kbps"
                            },
                            "protocol=ipv6": {
                                "filter_id": "800::800",
                                "delay": "10.0ms",
                                "rate": "500Kbps",
                                "delay-distro": "2.0ms"
                            }
                        }
                    }
                }"""
            )

            runner.run()
            print_test_result(expected=expected, actual=runner.stdout, error=runner.stderr)

            assert json.loads(runner.stdout) == json.loads(expected)

            delete_all_rules(tc_target)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import unicode_literals

import re

import pytest

from tcconfig._split_line_list import split_line_list


class Test_split_line_list(object):
    @pytest.mark.parametrize(
        ["value", "separator", "is_include_matched_line", "is_strip", "expected"],
        [
            [
                ["abcdefg", "ABCDEFG", "1234"],
                re.compile("DEFG$"),
                False,
                True,
                [["abcdefg"], ["1234"]],
            ],
            [
                ["abcdefg", "ABCDEFG", "ABCDEFG", "1234"],
                re.compile("DEFG$"),
                False,
                True,
                [["abcdefg"], ["1234"]],
            ],
            [
                ["ABCDEFG", "abcdefg", "ABCDEFG", "1234", "ABCDEFG"],
                re.compile("DEFG$"),
                False,
                True,
                [["abcdefg"], ["1234"]],
            ],
            [
                ["abcdefg", "ABCDEFG", "1234"],
                re.compile("DEFG$"),
                True,
                True,
                [["abcdefg"], ["ABCDEFG", "1234"]],
            ],
            [["a", "  ", "b", "c"], re.compile("^$"), False, True, [["a"], ["b", "c"]]],
            [["a", "  ", "b", "c"], re.compile("^$"), False, False, [["a", "  ", "b", "c"]]],
            [["a", "b", "c"], None, False, True, [["a", "b", "c"]]],
        ],
    )
    def test_normal(self, value, separator, is_include_matched_line, is_strip, expected):
        assert split_line_list(value, separator, is_include_matched_line, is_strip) == expected

    @pytest.mark.parametrize(
        ["value", "separator", "is_include_matched_line", "is_strip", "expected"],
        [
            [None, "", False, True, TypeError],
            [[1, 2, 3], re.compile(""), False, True, AttributeError],
            [[1, 2, 3], re.compile(""), False, False, TypeError],
        ],
    )
    def test_exception(self, value, separator, is_include_matched_line, is_strip, expected):
        with pytest.raises(expected):
            split_line_list(value, separator, is_include_matched_line, is_strip)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import division, print_function

import pytest
import simplejson as json
from subprocrunner import SubprocessRunner

from tcconfig._const import Tc
from tcconfig.parser.shaping_rule import TcShapingRuleParser
from tcconfig.traffic_control import delete_all_rules

from .common import NullLogger, print_test_result


@pytest.fixture
def device_value(request):
    return request.config.getoption("--device")


class Test_tcconfig(object):
    @pytest.mark.parametrize(["overwrite"], [[""], ["--overwrite"]])
    def test_config_file_smoke(self, tmpdir, device_value, overwrite):
        if device_value is None:
            pytest.skip("device option is null")

        p = tmpdir.join("tcconfig.json")
        config = (
            "{"
            + '"{:s}"'.format(device_value)
            + ": {"
            + """
        "outgoing": {
            "dst-network=192.168.0.10/32, dst-port=8080, protocol=ip": {
                "delay": "10.0ms",
                "loss": "0.01%",
                "rate": "250Kbps",
                "delay-distro": "2.0ms"
            },
            "src-port=1234, protocol=ip": {
                "delay": "50.0ms",
                "rate": "1Gbps"
            }
        },
        "incoming": {
            "dst-network=192.168.10.0/24, protocol=ip": {
                "corrupt": "0.02%",
                "rate": "1500Kbps"
            }
        }
    }
}
"""
        )
        print("[config]\n{}\n".format(config))
        p.write(config)

        for device_option in [device_value, "--device {}".format(device_value)]:
            delete_all_rules(device_value)
            command = " ".join([Tc.Command.TCSET, "--import-setting", str(p), overwrite])
            SubprocessRunner(command).run()

            parser = TcShapingRuleParser(
                device=device_value,
                ip_version=4,
                logger=NullLogger(),
                tc_command_output=None,
                is_parse_filter_id=False,
            )
            result = parser.get_tc_parameter()

            print_test_result(expected=config, actual=result)

            assert result == json.loads(config)

            delete_all_rules(device_value)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import

from simplesqlite.query import And, Where

from ._const import Tc, TcSubCommand, TrafficDirection
from ._network import is_anywhere_network
from .parser.shaping_rule import TcShapingRuleParser


class TcShapingRuleFinder(object):
    @property
    def _parser(self):
        self.__shaping_rule_parser.parse()

        return self.__shaping_rule_parser

    def __init__(self, logger, tc):
        self.__logger = logger
        self.__tc = tc
        self.__shaping_rule_parser = TcShapingRuleParser(
            device=self.__tc.device,
            ip_version=self.__tc.ip_version,
            tc_command_output=self.__tc.tc_command_output,
            logger=self.__logger,
        )

    def clear(self):
        self.__shaping_rule_parser.clear()

    def find_qdisc_handle(self, parent):
        return self._parser.con.fetch_value(
            select=Tc.Param.HANDLE,
            table_name=TcSubCommand.QDISC.value,
            where=Where(Tc.Param.PARENT, parent),
        )

    def find_filter_param(self):
        import simplesqlite

        where_list = self.__get_filter_conditions()
        where_query = And(where_list)
        table_name = TcSubCommand.FILTER.value
        self.__logger.debug("find filter param: table={}, where={}".format(table_name, where_query))

        try:
            result = self._parser.con.select_as_dict(
                columns=[Tc.Param.FILTER_ID, Tc.Param.PRIORITY, Tc.Param.PROTOCOL],
                table_name=table_name,
                where=where_query,
            )
        except simplesqlite.TableNotFoundError:
            return None

        if not result:
            self.__logger.debug(
                "find filter param: empty result (table={}, where={})".format(
                    table_name, where_query
                )
            )
            return None

        param = result[0]
        self.__logger.debug(
            "find filter param: result={}, table={}, where={}".format(
                param, table_name, where_query
            )
        )

        return param

    def find_parent(self):
        where_list = self.__get_filter_conditions()
        table_name = TcSubCommand.FILTER.value
        parent = self._parser.con.fetch_value(
            select=Tc.Param.FLOW_ID, table_name=table_name, where=And(where_list)
        )

        self.__logger.debug(
            "find parent: result={}, table={}, where={}".format(parent, table_name, where_list)
        )

        return parent

    def is_exist_rule(self):
        return self.find_parent() is not None

    def is_any_filter(self):
        num_records = self._parser.con.fetch_num_records(table_name=TcSubCommand.FILTER.value)

        return num_records and num_records > 0

    def is_empty_filter_condition(self):
        from typepy import is_null_string

        return all(
            [
                is_anywhere_network(self.__tc.dst_network, self.__tc.ip_version),
                is_anywhere_network(self.__tc.src_network, self.__tc.ip_version),
                is_null_string(self.__tc.dst_port),
                is_null_string(self.__tc.src_port),
            ]
        )

    def get_parsed_device(self):
        if self.__tc.direction == TrafficDirection.OUTGOING:
            device = self._parser.device
        elif self.__tc.direction == TrafficDirection.INCOMING:
            device = self._parser.ifb_device

        return device

    def get_filter_string(self):
        return ", ".join(
            [where.to_query() for where in self.__get_filter_conditions() if where.value]
        )

    def __get_filter_conditions(self):
        if self.__tc.direction == TrafficDirection.OUTGOING:
            device = self._parser.device
        elif self.__tc.direction == TrafficDirection.INCOMING:
            device = self._parser.ifb_device

        return [
            Where(Tc.Param.DEVICE, device),
            Where(Tc.Param.PROTOCOL, self.__tc.protocol),
            Where(Tc.Param.DST_NETWORK, self.__tc.dst_network),
            Where(Tc.Param.SRC_NETWORK, self.__tc.src_network),
            Where(Tc.Param.DST_PORT, self.__tc.dst_port),
            Where(Tc.Param.SRC_PORT, self.__tc.src_port),
        ]



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import itertools

import pytest
from allpairspy import AllPairs
from humanreadable import ParameterError

from tcconfig._const import ShapingAlgorithm, Tc, TrafficDirection
from tcconfig._netem_param import (
    MAX_CORRUPTION_RATE,
    MAX_PACKET_DUPLICATE_RATE,
    MAX_PACKET_LOSS_RATE,
    MAX_REORDERING_RATE,
    MIN_CORRUPTION_RATE,
    MIN_PACKET_DUPLICATE_RATE,
    MIN_PACKET_LOSS_RATE,
    MIN_REORDERING_RATE,
    NetemParameter,
)
from tcconfig.traffic_control import TrafficControl

from .common import is_invalid_param


MIN_VALID_PACKET_LOSS = 0.0000000232  # [%]


@pytest.fixture
def device_option(request):
    return request.config.getoption("--device")


@pytest.mark.parametrize(
    ["value"],
    [
        ["".join(opt_list)]
        for opt_list in AllPairs(
            [
                ["0.1", "+1.25", "25"],
                ["kbps", "Kbps", " Kibps", "mbps", "Mbps", " Mibps", "gbps", "Gbps", " Gibps"],
            ]
        )
    ],
)
def test_TrafficControl_validate_bandwidth_rate_normal(value):
    tc_obj = TrafficControl(
        "dummy",
        netem_param=NetemParameter("dummy", bandwidth_rate=value),
        direction=TrafficDirection.OUTGOING,
        shaping_algorithm=ShapingAlgorithm.HTB,
    )
    tc_obj.netem_param.validate_bandwidth_rate()


@pytest.mark.parametrize(
    ["value", "expected"],
    [
        ["".join(opt_list), ParameterError]
        for opt_list in AllPairs([["0.1", "1"], ["", "kb", "KB", "mb", "MB", "gb", "GB"]])
    ]
    + [["".join(value), ParameterError] for value in ("B", "K", "M", "G")]
    + [["0bps", ParameterError]],
)
def test_TrafficControl_validate_bandwidth_rate_exception_1(value, expected):
    with pytest.raises(expected):
        tc_obj = TrafficControl(
            "dummy",
            netem_param=NetemParameter(
                "dummy",
                bandwidth_rate=value,
                latency_time=Tc.ValueRange.LatencyTime.MIN,
                latency_distro_time=Tc.ValueRange.LatencyTime.MIN,
            ),
            direction=TrafficDirection.OUTGOING,
            shaping_algorithm=ShapingAlgorithm.HTB,
        )
        tc_obj.netem_param.validate_bandwidth_rate()


@pytest.mark.parametrize(
    ["value", "expected"],
    [
        ["".join(opt_list), ParameterError]
        for opt_list in itertools.product(["-1", "0", "0.0"], ["bps", "Kbps", "MKbps", "GKbps"])
    ],
)
def test_TrafficControl_validate_bandwidth_rate_exception_2(value, expected):
    with pytest.raises(expected):
        tc_obj = TrafficControl(
            "dummy",
            netem_param=NetemParameter(
                "dummy",
                bandwidth_rate=value,
                latency_time=Tc.ValueRange.LatencyTime.MIN,
                latency_distro_time=Tc.ValueRange.LatencyTime.MIN,
            ),
            shaping_algorithm=ShapingAlgorithm.HTB,
        )
        tc_obj.netem_param.validate_bandwidth_rate()


class Test_TrafficControl_validate(object):
    @pytest.mark.parametrize(
        [
            "rate",
            "direction",
            "delay",
            "delay_distro",
            "loss",
            "duplicate",
            "corrupt",
            "src_network",
            "exclude_src_network",
            "dst_network",
            "exclude_dst_network",
            "src_port",
            "exclude_src_port",
            "dst_port",
            "exclude_dst_port",
            "shaping_algorithm",
        ],
        [
            opt_list
            for opt_list in AllPairs(
                [
                    [None, "", "100Kbps", "0.5Mbps", "0.1Gbps"],  # rate
                    [TrafficDirection.OUTGOING],
                    [Tc.ValueRange.LatencyTime.MIN, Tc.ValueRange.LatencyTime.MAX],  # delay
                    [Tc.ValueRange.LatencyTime.MIN, Tc.ValueRange.LatencyTime.MAX],  # delay_distro
                    [
                        None,
                        MIN_PACKET_LOSS_RATE,
                        MIN_VALID_PACKET_LOSS,
                        MAX_PACKET_LOSS_RATE,
                    ],  # loss
                    [None, MIN_PACKET_DUPLICATE_RATE, MAX_PACKET_DUPLICATE_RATE],  # duplicate
                    [None, MIN_CORRUPTION_RATE, MAX_CORRUPTION_RATE],  # corrupt
                    [None, "192.168.0.1", "192.168.0.0/24"],  # src_network
                    [None, "192.168.0.1", "192.168.0.0/25"],  # exclude_src_network
                    [
                        None,
                        "",
                        "192.168.0.1",
                        "192.168.0.254",
                        "192.168.0.1/32",
                        "192.168.0.0/24",
                    ],  # dst_network
                    [None, "192.168.0.1", "192.168.0.0/25"],  # exclude_dst_network
                    [None, 65535],  # src_port
                    [None, 22],  # exclude_src_port
                    [None, 65535],  # dst_port
                    [None, 22],  # exclude_dst_port
                    [ShapingAlgorithm.HTB, ShapingAlgorithm.TBF],  # shaping_algorithm
                ],
                n=3,
            )
        ],
    )
    def test_normal(
        self,
        device_option,
        rate,
        direction,
        delay,
        delay_distro,
        loss,
        duplicate,
        corrupt,
        src_network,
        exclude_src_network,
        dst_network,
        exclude_dst_network,
        src_port,
        exclude_src_port,
        dst_port,
        exclude_dst_port,
        shaping_algorithm,
    ):
        if device_option is None:
            pytest.skip("device option is null")

        tc_obj = TrafficControl(
            device=device_option,
            direction=direction,
            netem_param=NetemParameter(
                device=device_option,
                bandwidth_rate=rate,
                latency_time=delay,
                latency_distro_time=delay_distro,
                packet_loss_rate=loss,
                packet_duplicate_rate=duplicate,
                corruption_rate=corrupt,
            ),
            src_network=src_network,
            exclude_src_network=exclude_src_network,
            dst_network=dst_network,
            exclude_dst_network=exclude_dst_network,
            src_port=src_port,
            exclude_src_port=exclude_src_port,
            dst_port=dst_port,
            exclude_dst_port=exclude_dst_port,
            is_enable_iptables=True,
            shaping_algorithm=shaping_algorithm,
        )

        if is_invalid_param(rate, delay, loss, duplicate, corrupt, reordering=None):
            with pytest.raises(ParameterError):
                tc_obj.validate()
        else:
            tc_obj.validate()

    @pytest.mark.parametrize(
        ["direction", "delay", "reordering"],
        [
            opt_list
            for opt_list in AllPairs(
                [
                    [TrafficDirection.OUTGOING],
                    ["0.1ms", Tc.ValueRange.LatencyTime.MAX],  # delay
                    [MIN_REORDERING_RATE, MAX_REORDERING_RATE],  # reordering
                ],
                n=2,
            )
        ],
    )
    def test_normal_reordering(self, device_option, direction, delay, reordering):
        if device_option is None:
            pytest.skip("device option is null")

        tc_obj = TrafficControl(
            device=device_option,
            netem_param=NetemParameter(
                device=device_option,
                latency_time=delay,
                latency_distro_time=Tc.ValueRange.LatencyTime.MIN,
                reordering_rate=reordering,
            ),
            direction=direction,
            shaping_algorithm=ShapingAlgorithm.HTB,
        )

        tc_obj.validate()

    @pytest.mark.parametrize(
        ["value", "expected"],
        [
            [{"latency_time": "-1ms"}, ParameterError],
            [{"latency_time": "61min"}, ParameterError],
            [{"latency_time": "100ms", "latency_distro_time": "-1ms"}, ParameterError],
            [{"latency_time": "100ms", "latency_distro_time": "61min"}, ParameterError],
            [{"packet_loss_rate": MIN_PACKET_LOSS_RATE - 0.1}, ParameterError],
            [{"packet_loss_rate": MAX_PACKET_LOSS_RATE + 0.1}, ParameterError],
            [
                {"latency_time": "100ms", "packet_duplicate_rate": MIN_PACKET_DUPLICATE_RATE - 0.1},
                ParameterError,
            ],
            [
                {"latency_time": "100ms", "packet_duplicate_rate": MAX_PACKET_DUPLICATE_RATE + 0.1},
                ParameterError,
            ],
            [{"corruption_rate": MIN_CORRUPTION_RATE - 0.1}, ParameterError],
            [{"corruption_rate": MAX_CORRUPTION_RATE + 0.1}, ParameterError],
            [{"reordering_rate": MIN_REORDERING_RATE - 0.1}, ParameterError],
            [{"reordering_rate": MAX_REORDERING_RATE + 0.1}, ParameterError],
            [{Tc.Param.DST_NETWORK: "192.168.0."}, ParameterError],
            [{Tc.Param.DST_NETWORK: "192.168.0.256"}, ParameterError],
            [{Tc.Param.DST_NETWORK: "192.168.0.0/0"}, ParameterError],
            [{Tc.Param.DST_NETWORK: "192.168.0.0/33"}, ParameterError],
            [{Tc.Param.DST_NETWORK: "192.168.0.2/24"}, ParameterError],
            [{Tc.Param.DST_NETWORK: "192.168.0.0000/24"}, ParameterError],
            [{"src_port": -1}, ParameterError],
            [{"src_port": 65536}, ParameterError],
            [{"dst_port": -1}, ParameterError],
            [{"dst_port": 65536}, ParameterError],
        ],
    )
    def test_exception(self, device_option, value, expected):
        if device_option is None:
            pytest.skip("device option is null")

        tc_obj = TrafficControl(
            device=device_option,
            netem_param=NetemParameter(
                device=device_option,
                bandwidth_rate=value.get("bandwidth_rate"),
                latency_time=value.get("latency_time", Tc.ValueRange.LatencyTime.MIN),
                latency_distro_time=value.get("latency_distro_time", Tc.ValueRange.LatencyTime.MIN),
                packet_loss_rate=value.get("packet_loss_rate"),
                packet_duplicate_rate=value.get("packet_duplicate_rate"),
                corruption_rate=value.get("corruption_rate"),
            ),
            dst_network=value.get(Tc.Param.DST_NETWORK),
            src_port=value.get("src_port"),
            dst_port=value.get("dst_port"),
            shaping_algorithm=ShapingAlgorithm.HTB,
        )

        with pytest.raises(expected):
            tc_obj.validate()


class Test_TrafficControl_ipv4(object):
    @pytest.mark.parametrize(
        ["network", "is_ipv6", "expected_ip_ver", "expected_protocol", "expected_protocol_match"],
        [
            ["192.168.0.1", False, 4, "ip", "ip"],
            ["192.168.0.0/24", False, 4, "ip", "ip"],
            ["::1", True, 6, "ipv6", "ip6"],
            ["2001:db00::0/24", True, 6, "ipv6", "ip6"],
        ],
    )
    def test_normal(
        self,
        device_option,
        network,
        is_ipv6,
        expected_ip_ver,
        expected_protocol,
        expected_protocol_match,
    ):
        if device_option is None:
            pytest.skip("device option is null")

        tc_obj = TrafficControl(
            device=device_option,
            netem_param=NetemParameter(
                device=device_option,
                latency_time=Tc.ValueRange.LatencyTime.MIN,
                latency_distro_time=Tc.ValueRange.LatencyTime.MIN,
            ),
            dst_network=network,
            is_ipv6=is_ipv6,
            shaping_algorithm=ShapingAlgorithm.HTB,
        )

        assert tc_obj.ip_version == expected_ip_ver
        assert tc_obj.protocol == expected_protocol
        assert tc_obj.protocol_match == expected_protocol_match



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, unicode_literals

import logbook
import simplesqlite
import subprocrunner


def _disable_logger(l):
    try:
        l.disable()
    except AttributeError:
        l.disabled = True  # to support Logbook<1.0.0


logger = logbook.Logger("tcconfig")
_disable_logger(logger)


def set_logger(is_enable):
    if is_enable != logger.disabled:
        # logger setting have not changed
        return

    if is_enable:
        try:
            logger.enable()
        except AttributeError:
            logger.disabled = False  # to support Logbook<1.0.0
    else:
        _disable_logger(logger)

    simplesqlite.set_logger(is_enable)
    subprocrunner.set_logger(is_enable)


def set_log_level(log_level):
    """
    Set logging level of this module. The module using
    `logbook <https://logbook.readthedocs.io/en/stable/>`__ module for logging.

    :param int log_level:
        One of the log level of the
        `logbook <https://logbook.readthedocs.io/en/stable/api/base.html>`__.
        Disabled logging if the ``log_level`` is ``logbook.NOTSET``.
    :raises LookupError: If ``log_level`` is an invalid value.
    """

    # validate log level
    logbook.get_level_name(log_level)

    if log_level == logger.level:
        return

    if log_level == logbook.NOTSET:
        set_logger(is_enable=False)
    else:
        set_logger(is_enable=True)

    logger.level = log_level
    simplesqlite.set_log_level(log_level)
    subprocrunner.set_log_level(log_level)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import, division

import pingparsing
import pytest
import typepy
from subprocrunner import SubprocessRunner

from tcconfig._const import Tc
from tcconfig.traffic_control import delete_all_rules

from .common import ASSERT_MARGIN, DEADLINE_TIME


@pytest.fixture
def device_option(request):
    return request.config.getoption("--device")


@pytest.fixture
def dst_host_option(request):
    return request.config.getoption("--dst-host")


@pytest.fixture
def dst_host_ex_option(request):
    return request.config.getoption("--dst-host-ex")


@pytest.fixture
def transmitter():
    transmitter = pingparsing.PingTransmitter()
    transmitter.ping_option = "-i 0.2 -q"
    transmitter.deadline = DEADLINE_TIME

    return transmitter


@pytest.fixture
def pingparser():
    return pingparsing.PingParsing()


class Test_tcset_two_network(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:

      python setup.py test --addopts "--device=<test device> \n
          --dst-host=<hostname/IP-addr> --dst-host-ex=<hostname/IP-addr>"

    These tests expected to execute in the following environment:
       - Linux w/ iputils-ping package
       - English locale (for parsing ping output)
    """

    @pytest.mark.parametrize(["shaping_algo"], [["htb"]])
    def test_network(
        self,
        device_option,
        dst_host_option,
        dst_host_ex_option,
        transmitter,
        pingparser,
        shaping_algo,
    ):
        if device_option is None:
            pytest.skip("device option is null")
        if any([typepy.is_null_string(dst_host_option), typepy.is_null_string(dst_host_ex_option)]):
            pytest.skip("destination host is null")

        for tc_target in [device_option, "--device {}".format(device_option)]:
            delete_all_rules(tc_target)
            delay = 100

            # tc to specific network ---
            command_list = [
                Tc.Command.TCSET,
                tc_target,
                "--delay {:d}ms".format(delay),
                "--dst-network {:s}".format(dst_host_ex_option),
                "--shaping-algo {:s}".format(shaping_algo),
            ]
            assert SubprocessRunner(" ".join(command_list)).run() == 0

            # w/o tc network ---
            transmitter.destination = dst_host_option
            without_tc_rtt_avg = pingparser.parse(transmitter.ping().stdout).rtt_avg

            # w/ tc network ---
            transmitter.destination = dst_host_ex_option
            with_tc_rtt_avg = pingparser.parse(transmitter.ping().stdout).rtt_avg

            # assertion ---
            rtt_diff = with_tc_rtt_avg - without_tc_rtt_avg
            assert rtt_diff > (delay * ASSERT_MARGIN)

            # finalize ---
            delete_all_rules(tc_target)
